import random

from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class RandomGeneMutationSelector(BaseGeneMutationSelector):
    """
    Selects and mutates genes in a chromosome using a random mutation strategy.
    """

    def __init__(
        self,
        gene_mutation_rate_determinator: BaseGeneMutationRateDeterminator,
        gene_mutator: BaseGeneMutator,
    ):
        super().__init__(gene_mutation_rate_determinator, gene_mutator)

    def _mutate_chromosome(self):
        if self.number_of_mutation_genes < 1:
            self.number_of_mutation_genes = 1
        genes_to_mutate = list(self.chromosome)
        random.shuffle(genes_to_mutate)
        number_of_mutation_genes = (
            self._gene_mutation_rate_determinator.get_number_of_mutation_genes(
                self.chromosome, self.number_of_mutation_genes
            )
        )
        if number_of_mutation_genes < 1:
            number_of_mutation_genes = 1
        var_num = random.randint(1, number_of_mutation_genes)
        for g in genes_to_mutate[:var_num]:
            self._mutate_gene(g)

        return var_num



import math

from gadapt.utils.ga_utils import (
    normally_distributed_random,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class NormalDistributionGeneMutator(BaseGeneMutator):
    """
    Generates random or normally distributed values.
    """

    def _make_mutated_value(self):
        return self._make_normally_distributed_random_value_until_changed()

    def _make_normally_distributed_random_value_until_changed(self):
        return self._execute_function_until_value_changed(
            self._make_normally_distributed_random_value
        )

    def _calculate_normal_distribution_standard_deviation(self):
        return 0.05

    def _make_normally_distributed_random_value(self):
        curr_value = self.gene.variable_value
        if math.isnan(curr_value):
            curr_value = self.gene.decision_variable.make_random_value()
        decision_variable_range = (
            self.gene.decision_variable.max_value
            - self.gene.decision_variable.min_value
        )
        mean = (
            curr_value - self.gene.decision_variable.min_value
        ) / decision_variable_range
        normal_distribution_random_value = normally_distributed_random(
            mean, self._calculate_normal_distribution_standard_deviation(), 0, 1
        )
        number_of_steps = round(
            (normal_distribution_random_value * decision_variable_range)
            / self.gene.decision_variable.step
        )
        return (
            self.gene.decision_variable.min_value
            + number_of_steps * self.gene.decision_variable.step
        )



import random
from typing import List

from gadapt.operations.sampling.base_sampling import T, BaseSampling


class RandomSampling(BaseSampling):
    """
    "Random" algorithm for extracting a sample from the population.
    """

    def _prepare_sample(self, lst: List[T]) -> List[T]:
        members_for_action = random.sample(lst, len(lst))
        return [
            m.set_rank_and_return(rank)
            for rank, m in enumerate(members_for_action[: self.max_num])
        ]



"""
Results for the genetic algorithm execution
"""

from typing import List

import gadapt.adapters.string_operation.ga_strings as ga_strings


class GAResults:
    def __init__(self) -> None:
        """
        Results class for the genetic algorithm execution
        """
        self._success = True
        self.result_values = {}
        self._messages: List[str] = []
        self._min_cost_per_generation = []

    def __str__(self) -> str:
        return ga_strings.results_to_string(self)

    @property
    def result_values(self):
        """
        The dictionary that contains variables' optimized values.
        The key of this dictionary is the sequence number of variable adding, and\
            also the argument index in the cost function.
        The value of this dictionary is the optimized value for the variable.
        """
        return self._result_values

    @result_values.setter
    def result_values(self, value):
        self._result_values = value

    @property
    def min_cost(self) -> float:
        """
        Minimum cost found
        """
        return self._min_cost

    @min_cost.setter
    def min_cost(self, value: float):
        self._min_cost = value

    @property
    def min_cost_per_generation(self) -> List[float]:
        """
        Minimal costs for each generation
        """
        return self._min_cost_per_generation

    @min_cost_per_generation.setter
    def min_cost_per_generation(self, value: List[float]):
        self._min_cost_per_generation = value

    @property
    def number_of_iterations(self) -> float:
        """
        Number of iterations (generations)
        """
        return self._number_of_iterations

    @number_of_iterations.setter
    def number_of_iterations(self, value: float):
        self._number_of_iterations = value

    @property
    def success(self) -> bool:
        """
        Indicates if genetic algorithm optimization succeded
        """
        return self._success

    @success.setter
    def success(self, value: bool):
        self._success = value

    @property
    def messages(self):
        """
        Messages from the optimization
        """
        return self._messages

    @messages.setter
    def messages(self, value):
        self._messages = value

    @property
    def message(self) -> str:
        """
        Messages from the optimization sublimed to one string
        """
        return self._get_message()

    def _get_message(self):
        return ga_strings.get_results_message(self)



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class NumberOfGenerationsExitChecker(BaseExitChecker):
    """
    Exit check based on the number of generations.
    The GA exits when the defined number of generations is reached.
    """

    def __init__(self, number_of_generations: float) -> None:
        super().__init__(1)
        self.number_of_generations = number_of_generations

    def _is_exit(self):
        if self.population.population_generation >= self.number_of_generations:
            return True
        return False



from typing import List, Tuple, Optional

import gadapt.ga_model.message_levels as message_levels
from gadapt.ga_model.chromosome import Chromosome
from gadapt.adapters.ga_logging.logging_settings import init_logging
from gadapt.factory.ga_factory import BaseGAFactory
from gadapt.ga_model.ga_options import GAOptions
from gadapt.ga_model.ga_results import GAResults
from gadapt.ga_model.population import Population


class GAExecutor:
    """
    Executor for the genetic algorithm

    Args:
        ga_options (GAOptions): Options for GA execution
        factory (GAFactory): Factory for objects creation
    """

    def __init__(self, ga_options: GAOptions, factory: BaseGAFactory) -> None:
        self.population: Optional[Population] = None
        if ga_options is not None:
            self.ga_options = ga_options
        self.factory = factory
        self.chromosome_mutator = self.factory.get_gene_mutation_selector()
        self.population_mutator = self.factory.get_chromosome_mutation_selector()
        self.exit_checker = self.factory.get_exit_checker()
        self.cost_finder = self.factory.get_cost_finder()
        self.population_immigrator = self.factory.get_population_immigrator()
        self.chromosome_immigrator = self.factory.get_chromosome_immigrator()
        self.selector = self.factory.get_parent_selector()
        self.crossover = self.factory.get_crossover()
        self.variable_updater = self.factory.get_variable_updater()
        self.population_updater = self.factory.get_population_updater()
        self.gene_mutator = self.factory.get_gene_mutator()

    def execute(self) -> GAResults:
        """
        Executes the genetic algorithm
        """
        self.population = Population(self.ga_options)
        if self.population is None:
            raise Exception("population object is None!")
        results = GAResults()
        try:
            init_logging(self.ga_options.logging)
        except Exception as ex:
            results.messages.append(
                (
                    message_levels.WARNING,
                    "Logging failed. Error message: {exc}".format(exc=str(ex)),
                )
            )
            self.ga_options.logging = False
        try:

            self.find_costs()
            while not self.exit():
                self.immigrate()
                self.mate()
                self.mutate()
                self.find_costs()
            if self.population.timeout_expired:
                results.messages.append((message_levels.WARNING, "Timeout expired!"))
            best_individual = self.population.best_individual
            results.min_cost = self.population.min_cost
            results.number_of_iterations = self.population.population_generation
            results.min_cost_per_generation = self.population.min_cost_per_generation
            for g in best_individual:
                results.result_values[g.decision_variable.variable_id] = (
                    g.variable_value
                )
        except Exception as ex:
            results.success = False
            results.messages.append((message_levels.ERROR, str(ex)))
        return results

    def exit(self) -> bool:
        """
        Check exit from the GA
        """
        if self.population is None:
            raise Exception("population object must not be None!")
        self.population.population_generation += 1
        return self.exit_checker.check(self.population)

    def find_costs(self):
        """
        Finds costs for chromosomes
        """
        self.population.previous_avg_cost = self.population.avg_cost
        self.population.previous_min_cost = self.population.min_cost
        self.cost_finder.find_costs(self.population)
        self.variable_updater.update_variables(self.population)
        self.population_updater.update_population(self.population)

    def immigrate(self):
        """
        Immigrates new chromosomes
        """
        self.population_immigrator.immigrate(self.population)

    def mate(self):
        """
        Mates chromosomes
        """
        chromosome_pairs = self.select_mates()
        self.crossover.mate(chromosome_pairs, self.population)

    def mutate(self):
        """
        Mutates chromosomes in the population
        """
        self.population_mutator.mutate(self.population)

    def select_mates(self) -> List[Tuple[Chromosome, Chromosome]]:
        """
        Selects mates for pairing
        """
        return self.selector.select_mates(self.population)



import math
from abc import ABC, abstractmethod
from typing import List

from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class BaseChromosomeMutationSelector(ABC):
    """
    Selects and mutates the chromosomes in the population based on a specified number of mutated chromosomes.
    """

    def __init__(
        self,
        chromosome_mutation_rate_determinator: BaseChromosomeMutationRateDeterminator,
        gene_mutation_selector: BaseGeneMutationSelector,
    ) -> None:
        """
        Base class for selecting mating chromosomes in population
        Args:
            chromosome_mutation_rate_determinator: chromosome mutation rate determinator
        """
        super().__init__()
        self._chromosome_mutation_rate_determinator = (
            chromosome_mutation_rate_determinator
        )
        self.number_of_mutation_chromosomes = -1
        self._gene_mutation_selector = gene_mutation_selector

    def mutate(self, population):
        """
        Mutates chromosomes in the population
        Args:
            population: Population to mutate
        """
        self.population = population
        max_number_of_mutated_chromosomes = (
            population.options.number_of_mutation_chromosomes
        )
        self.number_of_mutation_chromosomes = self._chromosome_mutation_rate_determinator.get_number_of_mutation_chromosomes(
            self.population, max_number_of_mutated_chromosomes
        )
        self._mutate_population()

    @abstractmethod
    def _mutate_population(self):
        pass

    def mutate_chromosome(self, c: Chromosome, number_of_mutated_chromosomes: int):
        self._gene_mutation_selector.mutate(c, number_of_mutated_chromosomes)

    def _get_unallocated_chromosomes(self, sort_key_function=None) -> List[Chromosome]:
        def unallocated_chromosomes_condition(c: Chromosome) -> bool:
            return (
                math.isnan(c.cost_value)
                and (not c.is_immigrant)
                and c.population_generation == self.population.population_generation
                and not c.is_mutated
            )

        lst = [c for c in self.population if (unallocated_chromosomes_condition(c))]
        if sort_key_function is not None:
            lst.sort(key=sort_key_function)
        return lst



import random

from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_selector import (
    BaseChromosomeMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class RandomChromosomeMutationSelector(BaseChromosomeMutationSelector):
    """
    Implements the mutation of chromosomes in a population based on a random selection of chromosomes
    """

    def __init__(
        self,
        chromosome_mutation_rate_determinator: BaseChromosomeMutationRateDeterminator,
        gene_mutation_selector: BaseGeneMutationSelector,
    ) -> None:
        super().__init__(chromosome_mutation_rate_determinator, gene_mutation_selector)

    def _mutate_population(self):
        if self.population is None:
            raise Exception("population must not be None")
        unallocated_chromosomes = self._get_unallocated_chromosomes(
            self._sort_key_random
        )
        mutation_chromosome_number = self.number_of_mutation_chromosomes
        if mutation_chromosome_number == 0:
            return 0
        chromosomes_for_mutation = unallocated_chromosomes[:mutation_chromosome_number]
        for c in chromosomes_for_mutation:
            self._gene_mutation_selector.mutate(
                c, self.population.options.number_of_mutation_genes
            )
        return mutation_chromosome_number

    def _sort_key_random(self, _: Chromosome):
        return random.random()



import random

from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_selector import (
    BaseChromosomeMutationSelector,
)
from gadapt.operations.sampling.base_sampling import BaseSampling
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class ParentDiversityChromosomeMutationSelector(BaseChromosomeMutationSelector):
    """
    Selects and mutates chromosomes in a population based on their parent diversity.
    """

    def __init__(
        self,
        chromosome_mutation_rate_determinator: BaseChromosomeMutationRateDeterminator,
        gene_mutation_selector: BaseGeneMutationSelector,
        sampling: BaseSampling,
    ) -> None:
        super().__init__(chromosome_mutation_rate_determinator, gene_mutation_selector)
        self._sampling = sampling

    def _sort_key_parent_diversity_random(self, c: Chromosome):
        return (c.parent_diversity, random.random())

    def _mutate_population(self):
        if self.population is None:
            raise Exception("Population must not be null")
        unallocated_chromosomes: list[Chromosome] = self._get_unallocated_chromosomes(
            self._sort_key_parent_diversity_random
        )
        chromosomes_for_mutation: list[Chromosome] = []
        if self.population.options.must_mutate_for_same_parents:
            chromosomes_for_mutation = [
                c for c in unallocated_chromosomes if c.parent_diversity == 0
            ]
        chromosomes_for_mutation_count = len(chromosomes_for_mutation)
        rest_number = (
            self.number_of_mutation_chromosomes - chromosomes_for_mutation_count
        )
        if rest_number > 0:
            if self.population.options.must_mutate_for_same_parents:
                other_chromosomes_for_mutation = [
                    c for c in unallocated_chromosomes if (not c.parent_diversity == 0)
                ]
            else:
                other_chromosomes_for_mutation = [c for c in unallocated_chromosomes]
            other_chromosomes_for_mutation = self._sampling.get_sample(
                other_chromosomes_for_mutation,
                rest_number,
                lambda c: c.parent_diversity,
            )
            chromosomes_for_mutation.extend(other_chromosomes_for_mutation)
        for c in chromosomes_for_mutation:
            self._gene_mutation_selector.mutate(
                c, self.population.options.number_of_mutation_genes
            )
        return len(chromosomes_for_mutation)



import gadapt.utils.ga_utils as ga_utils
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class CrossDiversityChromosomeMutationRateDeterminator(
    BaseChromosomeMutationRateDeterminator
):
    """
    Determining the number of chromosomes to be mutated in a population based on the cross-diversity coefficient of the decision variables.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_number_of_mutation_chromosomes(self) -> int:
        def get_mutation_rate() -> float:
            avg_rsd = ga_utils.average(
                [
                    dv.cross_diversity_coefficient
                    for dv in self.population.options.decision_variables
                ]
            )
            if avg_rsd > 1:
                avg_rsd = 1
            if avg_rsd < 0:
                avg_rsd = 0
            return 1 - avg_rsd

        mutation_rate = get_mutation_rate()
        f_return_value = mutation_rate * float(self.max_number_of_mutation_chromosomes)
        return round(f_return_value)



from abc import ABC, abstractmethod


class BaseChromosomeImmigrator(ABC):
    """
    Base class for chromosome immigration
    """

    def __init__(self):
        super().__init__()
        self.chromosome = None

    def immigrate(self, c):
        """
        Makes one chromosome immigrant
        Args:
            c: chromosome to become the immigrant
        """
        self.chromosome = c
        self._immigrate_chromosome()
        self._chromosome_immigrated()

    @abstractmethod
    def _immigrate_chromosome(self):
        pass

    def _chromosome_immigrated(self):
        self.chromosome.is_immigrant = True
        if self.chromosome.first_immigrant_generation == 0:
            self.chromosome.first_immigrant_generation += 1
        self.chromosome.last_immigrant_generation = 1
        self.chromosome.first_mutant_generation = 0
        self.chromosome.last_mutant_generation = 0
        self.chromosome.set_chromosome_string_none()



class ExecuteUntilChanges:

    def __init__(self, current_value, number_of_attempts=5):
        self.current_value = current_value
        self.number_of_attempts = number_of_attempts

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            return self._execute_function_until_value_changed(func)

        return wrapper

    def _execute_function_until_value_changed(self, f):
        i = 0
        while True:
            new_function_return_value = f()
            i += 1
            if (
                new_function_return_value != self.current_value
                or i >= self.number_of_attempts
            ):
                break
        return new_function_return_value



from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)
from gadapt.utils import ga_utils
from operations.crossover.blending_crossover import BlendingCrossover


class BlendingParentDiversityCrossover(BlendingCrossover):
    """
    Blending Crossover. Genes from parents' chromosomes are combined in a blending way.
    Calculates diversity of parents and save it to chromosomes.
    """

    def __init__(
        self,
        mutator: BaseGeneMutationSelector,
        immigrator: BaseChromosomeImmigrator,
    ):
        super(BlendingParentDiversityCrossover, self).__init__(mutator, immigrator)
        self._genetic_diversity = None

    def _get_genetic_diversity(self) -> float:
        return abs(
            self._mother_gene.variable_value - self._father_gene.variable_value
        ) / (
            self._father_gene.decision_variable.max_value
            - self._father_gene.decision_variable.min_value
        )

    def _get_parent_diversity(self):
        return round(ga_utils.average(self._genetic_diversity), 2)

    def _decision_variable_crossed(self):
        self._genetic_diversity.append(self._get_genetic_diversity())

    def _all_decision_variable_crossed(self):
        parent_diversity = self._get_parent_diversity()
        self._offspring1.parent_diversity = parent_diversity
        self._offspring2.parent_diversity = parent_diversity

    def _cross_genetic_material(self):
        self._genetic_diversity = []
        super()._cross_genetic_material()



import random
from typing import List

from gadapt.operations.sampling.base_sampling import T, BaseSampling


class RouletteWheelSampling(BaseSampling):
    """
    "RouletteWheel" (also known as "Weighted Random Pairing") algorithm for\
        extracting a sample from the population.
    """

    def _prepare_sample(self, lst: List[T]) -> List[T]:
        """
        Prepares a sample from a list of objects.
        It calculates the cumulative probabilities for each object based on their rank
        and uses these probabilities to select objects for the sample.
        """
        rank_sum = sum(range(1, len(lst) + 1))
        cumulative_probability_list: List[float] = []
        action_probability = 0.0
        n_keep = len(lst)
        for j in range(len(lst)):
            n = n_keep - j
            probability_for_action = float(n_keep - n + 1) / float(rank_sum)
            action_probability += probability_for_action
            cumulative_probability_list.append(action_probability)
        rank = 0
        unallocated_members = [rm for rm in lst]
        unallocated_members.sort(key=self._sort_key, reverse=True)
        members_for_action = []
        for j in range(self.max_num):
            if len(unallocated_members) == 0:
                continue
            for i_c_p_l, m in enumerate(unallocated_members):
                m.action_probability = cumulative_probability_list[i_c_p_l]
            rnd_value = random.random()
            max_prob = (
                max(unallocated_members, key=lambda m: m.action_probability)
            ).action_probability
            rnd_value = rnd_value * max_prob
            mem = next(
                mb for mb in unallocated_members if mb.action_probability >= rnd_value
            )
            if mem is not None:
                rank += 1
                mem.rank = rank
                members_for_action.append(mem)
            unallocated_members = [um for um in unallocated_members if um.rank == -1]
        members_for_action.sort(key=lambda g: g.rank)
        return members_for_action



import numpy as np

from gadapt.operations.mutation.gene_mutation.normal_distribution_gene_mutator import (
    NormalDistributionGeneMutator,
)


class NormalDistributionCrossDiversityGeneMutator(NormalDistributionGeneMutator):
    """
    Generates random or normally distributed values. Calculates standard deviation based on
    the cross-diversity coefficient
    """

    def _calculate_normal_distribution_standard_deviation(self):
        min_std_dev = 0.05
        max_std_dev = 0.5
        std_dev_range = max_std_dev - min_std_dev
        dv_rsd = np.clip(self.gene.decision_variable.cross_diversity_coefficient, 0, 1)
        return min_std_dev + (std_dev_range * dv_rsd)



"""
Genetic algorithm definitions
"""

RANDOM = "random"
COST_DIVERSITY = "cost_diversity"
PARENT_DIVERSITY = "parent_diversity"
CROSS_DIVERSITY = "cross_diversity"
TOURNAMENT = "tournament"
FROM_TOP_TO_BOTTOM = "from_top_to_bottom"
ROULETTE_WHEEL = "roulette_wheel"
UNIFORM = "uniform"
BLENDING = "blending"
AVG_COST = "avg_cost"
MIN_COST = "min_cost"
REQUESTED = "requested"
GENERATIONS = "generations"
AVG = "avg"
MIN = "min"
PARAM_SEPARATOR = ","
FLOAT_NAN = float("NaN")
NOT_IMPLEMENTED = "not implemented"
EXTREME_POINTED = "extreme_pointed"
NORMAL_DISTRIBUTION = "normal_distribution"
POPULATION_MUTATION_STRINGS = [
    COST_DIVERSITY,
    PARENT_DIVERSITY,
    RANDOM,
    CROSS_DIVERSITY,
]
POPULATION_MUTATION_SELECTION_STRINGS = [RANDOM, PARENT_DIVERSITY]
CHROMOSOME_MUTATION_STRINGS = [RANDOM, CROSS_DIVERSITY]
GENE_MUTATION_STRINGS = [RANDOM, NORMAL_DISTRIBUTION, CROSS_DIVERSITY, EXTREME_POINTED]
SELECTION_STRINGS = [RANDOM, FROM_TOP_TO_BOTTOM, ROULETTE_WHEEL, TOURNAMENT]
EXIT_CRITERIA_STRINGS = [AVG_COST, MIN_COST, REQUESTED, GENERATIONS]



import random
from typing import List

from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class ComposedChromosomeMutationRateDeterminator(
    BaseChromosomeMutationRateDeterminator
):
    """
    Allows for the composition of multiple determinators to be used in a random order.
    """

    def __init__(self) -> None:
        super().__init__()
        self.determinators: List[BaseChromosomeMutationRateDeterminator] = []

    def append(self, determinator: BaseChromosomeMutationRateDeterminator):
        """
        Appends determinator to the composition of determinators
        """
        self.determinators.append(determinator)

    def _get_number_of_mutation_chromosomes(self):
        if self.population is None:
            raise Exception("Population must not be null")
        if len(self.determinators) == 0:
            raise Exception("at least one mutator must be added")
        for determinator in self.determinators:
            determinator.population = self.population
            determinator.max_number_of_mutation_chromosomes = (
                self.max_number_of_mutation_chromosomes
            )
        if len(self.determinators) > 1:
            random.shuffle(self.determinators)
        current_determinator = self.determinators[0]
        return current_determinator._get_number_of_mutation_chromosomes()



from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)


class RandomChromosomeImmigrator(BaseChromosomeImmigrator):
    """
    New immigrated chromosome with random genes
    """

    def _immigrate_chromosome(self):
        for g in self.chromosome:
            g.set_random_value()



import random
from typing import List

from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class ComposedGeneMutator(BaseGeneMutator):

    def __init__(self) -> None:
        super().__init__()
        self.mutators: List[BaseGeneMutator] = []

    def append(self, mutator: BaseGeneMutator):
        """
        Appends gene mutator to a list of mutators.
        Args:
            mutator: An instance of the BaseGeneMutator class that will be added to the list of mutator.
        """
        self.mutators.append(mutator)

    def _make_mutated_value(self):
        if self.gene is None:
            raise Exception("Gene must not be null")
        if len(self.mutators) == 0:
            raise Exception("at least one mutator must be added")
        for mutator in self.mutators:
            mutator.gene = self.gene
        if len(self.mutators) > 1:
            random.shuffle(self.mutators)
        return self.mutators[0]._make_mutated_value()



import datetime
import logging
import os

from gadapt.utils.TimeStampFormatter import TimestampFormatter


def init_logging(is_logging: bool):
    """
    Initializes logging for genetic algorithm
    """
    logger = logging.getLogger("gadapt_logger")
    if not is_logging:
        logger.disabled = True
        return
    else:
        logger.disabled = False
    path = os.path.join(os.getcwd(), "log")
    if not os.path.exists(path):
        os.mkdir(path)
    now = datetime.datetime.now()

    formatted_date_time = (
        now.strftime("%Y_%m_%d_%H_%M_%S_") + f"{now.microsecond // 1000:03d}"
    )
    logpath = os.path.join(path, f"gadapt_log_{formatted_date_time}.log")
    handler = logging.FileHandler(logpath)
    handler.setFormatter(
        TimestampFormatter("%(asctime)s - %(levelname)s - %(message)s")
    )
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)


def gadapt_log_info(msg: str):
    logger = logging.getLogger("gadapt_logger")
    if logger.disabled:
        return
    try:
        logger.info(msg)
    except Exception:
        pass



import random
from functools import reduce
from typing import List

import numpy as np

"""
Genetic algorithm utility
"""


def get_rand_bool():
    """
    Returns ranfom boolean value
    """
    rand_int = random.getrandbits(1)
    return bool(rand_int)


def get_rand_bool_with_probability(probability: float):
    """
    Returns random value with passed probability for returned value to be True
    Args:
        probability (float): The probability for True value
    """
    r = random.uniform(0, 1)
    if r <= probability:
        return True
    return False


def average(lst: List):
    """
    Average value of elements in the list
    Args:
        lst: List for finding the average value
    """
    return reduce(lambda a, b: a + b, lst) / len(lst)


def try_get_int(s):
    """
    If "s" objects is converted to int, it returns converted value.
    Othervise, returns s value.
    Args:
        s: Object to convert to int
    """
    if isinstance(s, int):
        return s
    if isinstance(s, str):
        try:
            n = int(s)
            return n
        except Exception:
            return s
    return s


def try_get_float(s):
    """
    If "s" objects is converted to float, it returns converted value.
    Othervise, returns s value.
    Args:
        s: Object to convert to float
    """
    if isinstance(s, float):
        return s
    if isinstance(s, str) or isinstance(s, int):
        try:
            n = float(s)
            return n
        except Exception:
            return s
    return s


def try_get_bool(s):
    """
    If "s" objects is converted to bool, it returns converted value.
    Othervise, returns s value.
    Args:
        s: Object to convert to bool
    """
    if isinstance(s, bool):
        return s
    if isinstance(s, str) or isinstance(s, int):
        try:
            n = bool(s)
            return n
        except Exception:
            return s
    return s


def prepare_string(s: str):
    """
    Removes spaces and new lines from the string
    Args:
        s (str): string to remove spaces and new lines
    Returns:
        str: string without spaces and new lines
    """
    str_return = s
    while " " in str_return or "\n" in str_return:
        str_return = str_return.replace(" ", "").replace("\n", "")
    return str_return


def normally_distributed_random(mean, std_dev, lower_bound, upper_bound):
    num = np.random.normal(mean, std_dev)
    num = np.clip(num, lower_bound, upper_bound)
    return num


def average_difference(diff_list):
    if len(diff_list) < 2:
        return float("NaN")
    diff_list.sort()
    differences = [diff_list[i + 1] - diff_list[i] for i in range(len(diff_list) - 1)]
    avg_difference = float(sum(differences) / len(differences))
    return avg_difference



import random

from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)


class RandomGeneMutationRateDeterminator(BaseGeneMutationRateDeterminator):
    """
    Determines the random number of mutation genes in a chromosome between 1 and the maximum number of mutation genes specified.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_number_of_mutation_genes(self) -> int:
        return random.randint(1, self.max_number_of_mutation_genes)



import math
from abc import ABC, abstractmethod
from typing import Tuple, List

from gadapt.ga_model.gene import Gene
from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class BaseCrossover(ABC):
    """Base Crossover Class

    Args:
        mutator (BaseGeneMutationSelector): mutation algorithm to
        be passed to offspring chromosomes
        immigrator (BaseChromosomeImmigrator): immigration algorithm
        to be passed to offspring chromosomes
        should be mutated on both sides with the same probability
    """

    def __init__(
        self,
        mutator: BaseGeneMutationSelector,
        immigrator: BaseChromosomeImmigrator,
    ):
        self._mutator = mutator
        self._immigrator = immigrator
        self._current_gene_number = -1

    def mate(self, chromosome_pairs: List[Tuple[Chromosome, Chromosome]], population):
        """
        Returns list of chromosome pairs using parents' genetic material

        Args:
            chromosome_pairs (List[Tuple[Chromosome, Chromosome]]) : List of chromosome pairs for mating
            population: Population
        """
        for chromosome1, chromosome2 in chromosome_pairs:
            offspring1, offspring2 = self._mate_pair(
                chromosome1, chromosome2, population.population_generation
            )
            population.add_chromosomes((offspring1, offspring2))
        current_len = len(population)
        chromosome_surplus = current_len - population.options.population_size
        if chromosome_surplus > 0:
            sorted_by_cost_value = sorted(
                population, key=lambda chrom: chrom.cost_value, reverse=True
            )
            i = 0
            for c in sorted_by_cost_value:
                if i >= chromosome_surplus:
                    break
                if not math.isnan(c.cost_value):
                    population.chromosomes.remove(c)

    def _mate_pair(
        self, mother: Chromosome, father: Chromosome, population_generation: int
    ):
        """Returns two offspring chromosomes using parents' genetic material

        Args:
            mother (Chromosome): The first chromosome for mating
            father (Chromosome): The second chromosome for mating
            population_generation (int): Current generation in the population

        Returns:
            Chromosome: the first offspring chromosome
            Chromosome: the second offspring chromosome
        """
        if len(mother) != len(father):
            raise Exception("Mother and father must have the same number of genes!")
        self._mother = mother
        self._father = father
        self._offspring1 = Chromosome(population_generation)
        self._offspring2 = Chromosome(population_generation)
        self._cross_genetic_material()
        self._increase_generation()
        return self._offspring1, self._offspring2

    def _cross_genetic_material(self):
        number_of_genes = len(self._father)
        for self._current_gene_number in range(number_of_genes):
            self._mother_gene, self._father_gene = self._get_mother_father_genes()
            decision_variable_father = self._mother_gene.decision_variable
            decision_variable_mother = self._father_gene.decision_variable
            if decision_variable_father != decision_variable_mother:
                decision_variable_mother = next(
                    (
                        item.decision_variable
                        for item in self._mother
                        if item.decision_variable == decision_variable_father
                    ),
                    None,
                )
            if decision_variable_mother is None:
                raise Exception(
                    "chromosomes in crossover do not have the same structure!"
                )
            self._decision_variable_crossed()
            var1, var2 = self._combine()
            self._offspring1.add_gene(decision_variable_father, var1)
            self._offspring2.add_gene(decision_variable_father, var2)
        self._all_decision_variable_crossed()
        self._offspring1.mother_id = self._mother.chromosome_id
        self._offspring2.mother_id = self._mother.chromosome_id
        self._offspring1.father_id = self._father.chromosome_id
        self._offspring2.father_id = self._father.chromosome_id

    def _get_mother_father_genes(self) -> Tuple[Gene, Gene]:
        if self._current_gene_number == -1:
            raise Exception("_current_gene_number not set")
        father_gene = self._father[self._current_gene_number]
        mother_gene = self._mother[self._current_gene_number]
        return mother_gene, father_gene

    @abstractmethod
    def _combine(self):
        pass

    def _increase_generation(self):
        current_generation = self._mother.chromosome_generation
        if (
            current_generation == 0
            or current_generation < self._father.chromosome_generation
        ):
            current_generation = self._father.chromosome_generation
        current_generation += 1
        self._offspring1.chromosome_generation = current_generation
        self._offspring2.chromosome_generation = current_generation

        current_generation = 0
        if (
            self._mother.first_mutant_generation > 0
            or self._father.first_mutant_generation > 0
        ):
            current_generation = self._mother.first_mutant_generation
            if (
                current_generation == 0
                or self._father.first_mutant_generation > current_generation
            ):
                current_generation = self._father.first_mutant_generation
            current_generation += 1
        self._offspring1.first_mutant_generation = current_generation
        self._offspring2.first_mutant_generation = current_generation

        current_generation = 0
        if (
            self._mother.last_mutant_generation > 0
            or self._father.last_mutant_generation > 0
        ):
            current_generation = self._mother.last_mutant_generation
            if current_generation == 0 or (
                0 < self._father.last_mutant_generation < current_generation
            ):
                current_generation = self._father.last_mutant_generation
            current_generation += 1
        self._offspring1.last_mutant_generation = current_generation
        self._offspring2.last_mutant_generation = current_generation

        current_generation = 0
        if (
            self._mother.first_immigrant_generation > 0
            or self._father.first_immigrant_generation > 0
        ):
            current_generation = self._mother.first_immigrant_generation
            if (
                current_generation == 0
                or self._father.first_immigrant_generation > current_generation
            ):
                current_generation = self._father.first_immigrant_generation
            current_generation += 1
        self._offspring1.first_immigrant_generation = current_generation
        self._offspring2.first_immigrant_generation = current_generation

        current_generation = 0
        if (
            self._mother.last_immigrant_generation > 0
            or self._father.last_immigrant_generation > 0
        ):
            current_generation = self._mother.last_immigrant_generation
            if current_generation == 0 or (
                0 < self._father.last_immigrant_generation < current_generation
            ):
                current_generation = self._father.last_immigrant_generation
            current_generation += 1
        self._offspring1.last_immigrant_generation = current_generation
        self._offspring2.last_immigrant_generation = current_generation

    def _decision_variable_crossed(self):
        pass

    def _all_decision_variable_crossed(self):
        pass



from abc import ABC, abstractmethod

from gadapt.ga_model.gene import Gene
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class BaseGeneMutationSelector(ABC):
    """
    Provides a framework for selecting and mutating genes in a chromosome.
    """

    def __init__(
        self,
        gene_mutation_rate_determinator: BaseGeneMutationRateDeterminator,
        gene_mutator: BaseGeneMutator,
    ):
        """
        Initializes an instance of the class by setting the gene_mutation_rate_determinator attribute.
        """
        self.chromosome = None
        self._gene_mutation_rate_determinator = gene_mutation_rate_determinator
        self._gene_mutator = gene_mutator
        self.number_of_mutation_genes = -1

    def mutate(self, c, number_of_mutation_genes: int):
        """
        Selects and mutates genes in the chromosome
        Args:
            c: Chromosome to mutate
            number_of_mutation_genes: Number of mutation genes
        """
        self.chromosome = c
        self.number_of_mutation_genes = number_of_mutation_genes
        self._before_mutated()
        self._mutate_chromosome()
        self._chromosome_mutated()

    @abstractmethod
    def _mutate_chromosome(self):
        pass

    def _mutate_gene(self, g: Gene):
        self._gene_mutator.mutate(g)
        self._gene_mutated(g)

    def _gene_mutated(self, g):
        self.chromosome.mutated_variables_id_list.append(
            g.decision_variable.variable_id
        )

    def _chromosome_mutated(self):
        self.chromosome.is_mutated = True
        if self.chromosome.first_mutant_generation == 0:
            self.chromosome.first_mutant_generation += 1
        self.chromosome.last_mutant_generation = 1

    def _before_mutated(self):
        self.chromosome.mutated_variables_id_list.clear()



from typing import List

from gadapt.operations.sampling.base_sampling import T, BaseSampling


class FromTopToBottomSampling(BaseSampling):
    """
    "From Top To Bottom" algorithm for extracting a sample from the population.
    """

    def _prepare_sample(self, lst: List[T]) -> List[T]:
        members_for_action = sorted(lst, key=self._sort_key)
        return [
            m.set_rank_and_return(rank)
            for rank, m in enumerate(members_for_action[: self.max_num])
        ]



from abc import ABC, abstractmethod


class BaseVariableUpdater(ABC):
    """
    Base class for variable update
    """

    def __init__(self):
        self.population = None

    def update_variables(self, population):
        self.population = population
        self._update_variables()

    @abstractmethod
    def _update_variables(self):
        pass



import statistics as stat

from gadapt.ga_model.decision_variable import DecisionVariable
from gadapt.operations.variable_update.base_variable_updater import BaseVariableUpdater


class CommonVariableUpdater(BaseVariableUpdater):
    """
    Common variable updater
    """

    def _update_variables(self):

        def scale_values(dv: DecisionVariable, values):
            scaled_values = []
            if dv.min_value == dv.max_value:
                return [0.5] * len(
                    values
                )  # If min_val and max_val are the same, return a list of 0.5s

            for value in values:
                scaled_value = (value - dv.min_value) / (dv.max_value - dv.min_value)
                scaled_values.append(scaled_value)
            return scaled_values

        unique_values_per_variables = {}
        values_per_variables = {}
        for c in self.population:
            if c.is_immigrant:
                continue
            for g in c:
                unique_var_values = unique_values_per_variables.get(
                    g.decision_variable, None
                )
                var_values = values_per_variables.get(g.decision_variable, None)
                if unique_var_values is None:
                    unique_var_values = set()
                    unique_values_per_variables[g.decision_variable] = unique_var_values
                if var_values is None:
                    var_values = []
                    values_per_variables[g.decision_variable] = var_values
                unique_var_values.add(g.variable_value)
                var_values.append(g.variable_value)
        for key in unique_values_per_variables:
            if len(unique_values_per_variables[key]) == 1:
                key.stacked = True
            else:
                key.stacked = False
        for key in values_per_variables:
            if key.stacked:
                key.cross_diversity_coefficient = 0.0
                continue
            values_scaled = scale_values(key, values_per_variables[key])
            st_dev = stat.stdev(values_scaled)
            if key.initial_st_dev < 0:
                key.initial_st_dev = st_dev
            if max(values_scaled) - min(values_scaled) == 0:
                key.cross_diversity_coefficient = 0.0
            else:
                key.cross_diversity_coefficient = st_dev / key.initial_st_dev



from abc import ABC, abstractmethod


class BasePopulationUpdater(ABC):
    """
    Base class for population update
    """

    def __init__(self):
        super().__init__()
        self.population = None

    def update_population(self, population):
        self.population = population
        self._update_population()

    @abstractmethod
    def _update_population(self):
        pass



import random

from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class RandomChromosomeMutationRateDeterminator(BaseChromosomeMutationRateDeterminator):
    """
    Provides a random method for determining the number of chromosomes to be mutated in a population.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_number_of_mutation_chromosomes(self) -> int:
        return random.randint(1, self.max_number_of_mutation_chromosomes)



from abc import ABC, abstractmethod
from typing import List

import gadapt.ga_model.message_levels as message_levels


class BaseOptionsValidator(ABC):
    def __init__(self, options) -> None:
        """
        Base class for options validation
        Args:
            options: Options to validate
        """
        self._validation_messages: List[str] = []
        self.options = options
        self.success = True

    def validate(self) -> None:
        """
        Validates options
        """
        self._validate_options()

    @abstractmethod
    def _validate_options(self) -> None:
        pass

    @property
    def validation_messages(self):
        """
        Returns validation messages
        """
        return self._validation_messages

    def _add_message(self, message, message_level=message_levels.ERROR):
        self.validation_messages.append((message_level, message))



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class RequestedCostExitChecker(BaseExitChecker):
    """
    Exit check based on requested cost.
    The GA exits when the minimum cost reaches a defined value.
    """

    def __init__(self, requested_cost: float) -> None:
        super().__init__(1)
        self.requested_cost = requested_cost

    def _is_exit(self):
        if self.population.min_cost <= self.requested_cost:
            return True
        return False



import random
from abc import ABC, abstractmethod
from typing import List, Tuple

from gadapt.ga_model.chromosome import Chromosome


class BaseParentSelector(ABC):
    """
    Base Parent Selector

    Selects individuals for mating from the population
    """

    def __init__(self):
        super().__init__()
        self.population = None

    def select_mates(self, population) -> List[Tuple[Chromosome, Chromosome]]:
        """
        Selects and returns individuals for the crossover from the population
        Args:
            population: the population for the mates selection
        """
        self.population = population
        return self._select_mates_from_population()

    def _select_mates_from_population(self) -> List[Tuple[Chromosome, Chromosome]]:
        working_chromosomes = self._select_potential_parents()
        num_of_crossover_parents = self.population.options.number_of_crossover_parents
        working_chromosomes_len = len(working_chromosomes)
        if num_of_crossover_parents > working_chromosomes_len:
            num_of_crossover_parents = working_chromosomes_len
        working_chromosomes = working_chromosomes[:num_of_crossover_parents]
        working_chromosomes_len = len(working_chromosomes)
        random_number_for_reserve = random.randint(0, working_chromosomes_len - 1)
        reserve_chromosome = working_chromosomes[random_number_for_reserve]
        list_of_mates: List[Tuple[Chromosome, Chromosome]] = []
        while len(working_chromosomes) > 0:
            c1 = working_chromosomes.pop(0)
            if len(working_chromosomes) > 0:
                c2 = working_chromosomes.pop(0)
            else:
                c2 = reserve_chromosome
            list_of_mates.append((c1, c2))
        return list_of_mates

    @abstractmethod
    def _select_potential_parents(self):
        pass



import random
from typing import List

from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_selector import (
    BaseChromosomeMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class ComposedChromosomeMutationSelector(BaseChromosomeMutationSelector):
    def __init__(
        self,
        chromosome_mutation_rate_determinator: BaseChromosomeMutationRateDeterminator,
        gene_mutation_selector: BaseGeneMutationSelector,
    ) -> None:
        super().__init__(chromosome_mutation_rate_determinator, gene_mutation_selector)
        self.selectors: List[BaseChromosomeMutationSelector] = []

    def append(self, selector: BaseChromosomeMutationSelector):
        """
        Appends selector to the composition of selectors
        """
        self.selectors.append(selector)

    def _mutate_population(self):
        if self.population is None:
            raise Exception("Population must not be null")
        if len(self.selectors) == 0:
            raise Exception("at least one mutator must be added")
        if len(self.selectors) > 1:
            random.shuffle(self.selectors)
        nmc = 0
        limit_number_of_mutation_chromosomes = self.number_of_mutation_chromosomes
        if limit_number_of_mutation_chromosomes == 0:
            return 0
        for m in self.selectors:
            if nmc < limit_number_of_mutation_chromosomes:
                m.population = self.population
                m.number_of_mutation_chromosomes = (
                    limit_number_of_mutation_chromosomes - nmc
                )
                mc = m._mutate_population()
                nmc += mc
        return nmc



from abc import ABC, abstractmethod
from typing import Optional

from gadapt.ga_model.population import Population
from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)


class BasePopulationImmigrator(ABC):
    """
    Base class for population immigration
    """

    def __init__(self, chromosome_immigrator: BaseChromosomeImmigrator):
        self.population: Optional[Population] = None
        self._chromosome_immigrator = chromosome_immigrator

    def immigrate(self, population):
        """
        Immigrates chromosomes into the population
        Args:
            population: Population to immigrate new chromosomes
        """
        self.population = population
        self._immigrate_population()

    @abstractmethod
    def _immigrate_population(self):
        pass

    def _immigrate_chromosome(self, c: Chromosome):
        self._chromosome_immigrator.immigrate(c)



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class MinCostExitChecker(BaseExitChecker):
    """
    Exit check based on minimal cost.
    The GA exits when there is no improvement in the minimal cost in a
    defined number of iterations.
    """

    def _is_exit(self):
        if self.population is None:
            raise Exception("population must not be null")
        return self.population.min_cost >= self.population.previous_min_cost



from abc import ABC, abstractmethod
from typing import List
from typing import TypeVar

from gadapt.ga_model.ranking_model import RankingModel

T = TypeVar("T", bound=RankingModel)


class BaseSampling(ABC):
    """
    The algorithm for extracting a sample from the population.
    """

    def get_sample(self, lst: List[T], max_num, sort_key=None) -> List[T]:
        """
        Gets the sample from the list of RankingModel objects.
        Args:
            lst (List[RankingModel]): list of RankingModel objects to get the sample
            max_num: (int): Maximal number of objects in the sample
            sort_key: sorting key for the sample
        """
        if len(lst) == 0:
            return lst
        for m in lst:
            m.reset_for_sampling()
        if max_num < 1 or max_num > len(lst):
            self.max_num = len(lst)
        else:
            self.max_num = max_num
        self._sort_key = sort_key
        return self._prepare_sample(lst)

    @abstractmethod
    def _prepare_sample(self, lst: List[T]) -> List[T]:
        pass



"""
Decision variable
"""

import random
import sys

import gadapt.ga_model.definitions as definitions


class DecisionVariable:
    def __init__(self, id: int) -> None:
        """
        Decision variable class defines genes.
        Each gene has a reference to one decision variable.
        Decision variable contains common values for genes: variable id, maximal\
            value, minimal value, step.
        Args:
            id (int): identifier of the decision variable
        """
        self._max_value = sys.float_info.min
        self._decimal_places = -1
        self._stacked = False
        self.variable_id = id
        self._standard_deviation = definitions.FLOAT_NAN
        self._initial_st_dev = -1.0

    def __eq__(self, other):
        if not isinstance(other, DecisionVariable):
            return False
        return self.variable_id == other.variable_id

    def __hash__(self) -> int:
        return self.variable_id

    @property
    def variable_id(self) -> int:
        """
        Unique ID for decision variable
        """
        return self._variable_id

    @variable_id.setter
    def variable_id(self, value: int):
        self._variable_id = value

    @property
    def max_value(self) -> float:
        """
        Max gene value
        """
        return self._max_value

    @max_value.setter
    def max_value(self, value: float):
        self._max_value = value

    @property
    def min_value(self) -> float:
        """
        Min gene value
        """
        return self._min_value

    @min_value.setter
    def min_value(self, value: float):
        self._min_value = value

    @property
    def step(self) -> float:
        """
        Optimization step
        """
        return self._step

    @step.setter
    def step(self, value: float):
        self._decimal_places = self._get_decimal_places(value)
        self._step = value

    def _get_decimal_places(self, num):
        num_str = str(num)
        if "e-" in num_str:
            num_str = num_str.split("e-")[-1]
            return int(num_str)
        if "." in num_str:
            _, fractional_part = num_str.split(".")
            return len(fractional_part)
        else:
            return 0

    @property
    def decimal_places(self) -> int:
        """
        Number of decimal places of the gene value
        """
        return self._decimal_places

    @property
    def stacked(self) -> bool:
        """
        Indicates if all genes have the same value for the same decision variable
        """
        return self._stacked

    @stacked.setter
    def stacked(self, value: bool):
        self._stacked = value

    @property
    def cross_diversity_coefficient(self) -> float:
        """
        Relative standard deviation of all genes for the same decision variable
        """
        return self._standard_deviation

    @cross_diversity_coefficient.setter
    def cross_diversity_coefficient(self, value: float):
        self._standard_deviation = value

    def make_random_value(self):
        """
        Makes random value, based on min value, max value, and step
        """
        number_of_steps = random.randint(
            0, round((self.max_value - self.min_value) / self.step)
        )
        return self.min_value + number_of_steps * self.step

    @property
    def initial_st_dev(self) -> float:
        return self._initial_st_dev

    @initial_st_dev.setter
    def initial_st_dev(self, value: float):
        self._initial_st_dev = value



from abc import ABC, abstractmethod

from gadapt.ga_model.gene import Gene


class BaseGeneMutator(ABC):
    """
    Mutates the variable value of a gene.
    """

    def mutate(self, g: Gene):
        self.gene = g
        self.gene.variable_value = self._make_mutated_value()

    @abstractmethod
    def _make_mutated_value(self):
        pass

    def _execute_function_until_value_changed(self, f):
        current_gene_value = self.gene.variable_value
        number_of_attempts = 5
        i = 0
        while True:
            new_gene_value = f()
            i += 1
            if new_gene_value != current_gene_value or i >= number_of_attempts:
                break
        return new_gene_value



import math

"""
Genetic algorithm string operations
"""


def gene_to_string(g):
    """
    Creates string  from the gene
    Args:
        g: Gene for string representation
    """
    return str(g.decision_variable.variable_id) + ": " + str(round(g.variable_value, 2))


def chromosome_to_string(c):
    """
    Creates string  from the chromosome
    Args:
        c: Chromosome for string representation
    """
    str_res = "Chromosome "
    id_str = str(c.chromosome_id)
    while len(id_str) < 4:
        id_str = " " + id_str
    str_res = str_res + id_str + " - "
    for g in c:
        str_res += (
            str(g.decision_variable.variable_id)
            + ": "
            + str(round(g.variable_value, 2))
            + "; "
        )
    str_res += "Cost value: " + str(c.cost_value) + "; "
    str_res += "Chromosome generation: " + str(c.chromosome_generation) + "; "
    if not (
        c.mother_id == -1
        or c.father_id == -1
        or c.mother_id is None
        or c.father_id is None
    ):
        str_res += (
            "Mother: " + str(c.mother_id) + ", Father: " + str(c.father_id) + "; "
        )
    if c.is_mutated and len(c.mutated_variables_id_list) > 0:
        str_res += "Chromosome mutated. Mutated variables: "
        for mv in c.mutated_variables_id_list:
            str_res += str(mv) + ", "
    if c.first_mutant_generation > 0:
        str_res += "First mutant generation: " + str(c.first_mutant_generation) + "; "
    if c.last_mutant_generation > 0:
        str_res += "Last mutant generation: " + str(c.last_mutant_generation) + "; "
    if c.is_immigrant:
        str_res += "Chromosome immigrated. "
    if c.first_immigrant_generation > 0:
        str_res += (
            "First immigrant generation: " + str(c.first_immigrant_generation) + "; "
        )
    if c.last_immigrant_generation > 0:
        str_res += (
            "Last immigrant generation: " + str(c.last_immigrant_generation) + "; "
        )
    return str_res


def population_to_string_list(p):
    """
    Creates string list from the population
    Args:
        p: Population for string list representation
    """
    str_list = []
    population_for_log = [c for c in p if not math.isnan(c.cost_value)]
    population_for_log.sort(key=lambda c: c.cost_value)
    for c in population_for_log:
        str_list.append(str(c))
    return str_list


def population_to_string(p):
    """
    Creates string from the population
    Args:
        gp Population for string representation
    """
    str_res = ""
    new_line = "\n"
    str_res += "Population number: " + str(p.population_generation) + new_line
    str_res += (
        "Min Cost: "
        + str(round(p.min_cost, 2))
        + "; Avg cost: "
        + str(round(p.avg_cost, 2))
        + "; "
    )
    str_res += "Best individual variable values: " + str(p.best_individual)
    str_list = population_to_string_list(p)
    for s in str_list:
        str_res += new_line + s
    return str_res


def results_to_string(r):
    """
    Creates string from the results
    Args:
        r: Results for string representation
    """
    mess = r._get_message()
    mess_exists = not (mess is None or len(mess) == 0)
    if r.success:
        rslt = "Min cost: " + str(r.min_cost)
        rslt += "\nNumber of iterations: " + str(r.number_of_iterations)
        rslt += "\nParameter values:"
        for x in r.result_values:
            rslt += "\n" + str(x) + ": " + str(r.result_values[x])
        if mess_exists:
            rslt += "\n" + mess
    elif not mess_exists:
        rslt = "\nCalculation not succesful."
    else:
        rslt = "\n" + mess
    return rslt


def get_results_message(r):
    """
    Creates a string from the list of messages in the results
    Args:
        r: Results containing the list of messages
    """
    message = ""
    new_line = "\n"
    for m in r.messages:
        message += m[0] + ": " + m[1] + new_line
    if message.endswith(new_line):
        message = message[: -len(new_line)]
    return message



from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)


class StrictGeneMutationRateDeterminator(BaseGeneMutationRateDeterminator):
    """
    Determines the strict number of mutation genes in a chromosome.
    """

    def __init__(self) -> None:
        super().__init__()

    def _get_number_of_mutation_genes(self) -> int:
        return self.max_number_of_mutation_genes



from typing import Tuple, List

import gadapt.ga_model.definitions as definitions
from gadapt.factory.ga_base_factory import BaseGAFactory
from gadapt.operations.cost_finding.base_cost_finder import BaseCostFinder
from gadapt.operations.cost_finding.elitism_cost_finder import ElitismCostFinder
from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.operations.crossover.blending_parent_diversity_crossover import (
    BlendingParentDiversityCrossover,
)
from gadapt.operations.exit_check.avg_cost_exit_checker import AvgCostExitChecker
from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker
from gadapt.operations.exit_check.min_cost_exit_checker import MinCostExitChecker
from gadapt.operations.exit_check.requested_cost_exit_checker import (
    RequestedCostExitChecker,
)
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.immigration.chromosome_immigration.random_chromosome_immigrator import (
    RandomChromosomeImmigrator,
)
from gadapt.operations.immigration.population_immigration.base_population_immigrator import (
    BasePopulationImmigrator,
)
from gadapt.operations.immigration.population_immigration.common_population_immigrator import (
    CommonPopulationImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.composed_gene_mutation_rate_determinator import (
    ComposedGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.composed_gene_mutation_selector import (
    ComposedGeneMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.cross_diversity_gene_mutation_rate_determinator import (
    CrossDiversityGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.cross_diversity_gene_mutation_selector import (
    CrossDiversityGeneMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.random_gene_mutation_rate_determinator import (
    RandomGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.random_gene_mutation_selector import (
    RandomGeneMutationSelector,
)
from gadapt.operations.mutation.chromosome_mutation.strict_gene_mutation_rate_determinator import (
    StrictGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator
from gadapt.operations.mutation.gene_mutation.extreme_pointed_gene_mutator import (
    ExtremePointedGeneMutator,
)
from gadapt.operations.mutation.gene_mutation.normal_distribution_gene_mutator import (
    NormalDistributionGeneMutator,
)
from gadapt.operations.mutation.gene_mutation.random_gene_mutator import (
    RandomGeneMutator,
)
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_selector import (
    BaseChromosomeMutationSelector,
)
from gadapt.operations.mutation.population_mutation.composed_chromosome_mutation_rate_determinator import (
    ComposedChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.composed_chromosome_mutation_selector import (
    ComposedChromosomeMutationSelector,
)
from gadapt.operations.mutation.population_mutation.cost_diversity_chromosome_mutation_rate_determinator import (
    CostDiversityChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.cross_diversity_chromosome_mutation_rate_determinator import (
    CrossDiversityChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.parent_diversity_chromosome_mutation_selector import (
    ParentDiversityChromosomeMutationSelector,
)
from gadapt.operations.mutation.population_mutation.random_chromosome_mutation_rate_determinator import (
    RandomChromosomeMutationRateDeterminator,
)
from gadapt.operations.mutation.population_mutation.random_chromosome_mutation_selector import (
    RandomChromosomeMutationSelector,
)
from gadapt.operations.mutation.population_mutation.strict_chromosome_mutation_rate_determinator import (
    StrictChromosomeMutationRateDeterminator,
)
from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.parent_selection.sampling_parent_selector import (
    SamplingParentSelector,
)
from gadapt.operations.sampling.base_sampling import BaseSampling
from gadapt.operations.sampling.from_top_to_bottom_sampling import (
    FromTopToBottomSampling,
)
from gadapt.operations.sampling.random_sampling import RandomSampling
from gadapt.operations.sampling.roulette_wheel_sampling import RouletteWheelSampling
from gadapt.operations.sampling.tournament_sampling import TournamentSampling
from gadapt.operations.variable_update.common_variable_updater import (
    CommonVariableUpdater,
)
from gadapt.operations.crossover.blending_crossover import BlendingCrossover
from gadapt.operations.crossover.uniform_crossover import UniformCrossover
from gadapt.operations.exit_check.number_of_generations_exit_checker import (
    NumberOfGenerationsExitChecker,
)
from gadapt.operations.mutation.gene_mutation.composed_gene_mutator import (
    ComposedGeneMutator,
)
from gadapt.operations.mutation.gene_mutation.normal_distribution_cross_diversity_gene_mutator import (
    NormalDistributionCrossDiversityGeneMutator,
)
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)
from gadapt.operations.population_update.common_population_updater import (
    CommonPopulationUpdater,
)


class GAFactory(BaseGAFactory):
    """
    Factory implementatiopn for creating  class instances based on GA options
    """

    def __init__(self):
        super().__init__()
        self.gene_mutation_rate_determinator = None
        self.chromosome_mutation_rate_determinator = None

    def _get_cost_finder(self) -> BaseCostFinder:
        """
        Cost Finder instance
        """
        return ElitismCostFinder()

    def _get_population_immigrator(self) -> BasePopulationImmigrator:
        """
        Population Immigrator Instance
        """
        return CommonPopulationImmigrator(self.get_chromosome_immigrator())

    def _get_chromosome_immigrator(self) -> BaseChromosomeImmigrator:
        """
        Chromosome Immigrator Instance
        """
        return RandomChromosomeImmigrator()

    def _get_gene_mutation_selector(self) -> BaseGeneMutationSelector:
        """
        Gene Mutation Selector Instance
        """
        return self._get_gene_mutation_selector_combined()

    def _get_gene_mutator_combined(self) -> BaseGeneMutator:
        if self._ga is None:
            raise Exception("_ga object must not be None!")
        mutator_strings = [
            gm.strip()
            for gm in self._ga.gene_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        gene_mutators: List[BaseGeneMutator] = []
        if definitions.RANDOM in mutator_strings:
            gene_mutators.append(RandomGeneMutator())
        if definitions.EXTREME_POINTED in mutator_strings:
            gene_mutators.append(ExtremePointedGeneMutator())
        if definitions.CROSS_DIVERSITY in mutator_strings:
            gene_mutators.append(NormalDistributionCrossDiversityGeneMutator())
        elif definitions.NORMAL_DISTRIBUTION in mutator_strings:
            gene_mutators.append(NormalDistributionGeneMutator())
        if len(gene_mutators) == 0:
            gene_mutators.append(NormalDistributionGeneMutator())
            gene_mutators.append(RandomGeneMutator())
        if len(gene_mutators) == 1:
            return gene_mutators[0]
        else:
            main_gene_mutator = ComposedGeneMutator()
            for mutator in gene_mutators:
                main_gene_mutator.append(mutator)
        return main_gene_mutator

    def _get_gene_mutator(self) -> BaseGeneMutator:
        """
        Chromosome Mutator Instance
        """
        return self._get_gene_mutator_combined()

    def _population_mutation_options_validation(self):
        """
        Validates population mutator options
        """
        mutator_strings = self._ga.population_mutation.split(
            definitions.PARAM_SEPARATOR
        )
        for s in mutator_strings:
            if s.strip() not in definitions.POPULATION_MUTATION_STRINGS:
                raise Exception(s + " is not defined as option for population mutation")

    def _make_chromosome_mutation_selector(
        self, population_mutator_string=None
    ) -> BaseChromosomeMutationSelector:
        """
        Population Mutator Instance
        """
        self._population_mutation_options_validation()
        return self._get_chromosome_mutation_selector_combined()

    def _get_chromosome_mutation_selector(self) -> BaseChromosomeMutationSelector:
        """
        Population Mutator Instance
        """
        return self._make_chromosome_mutation_selector()

    def _get_chromosome_mutation_rate_determinators(
        self,
    ) -> Tuple[
        BaseChromosomeMutationRateDeterminator, BaseChromosomeMutationRateDeterminator
    ]:
        if self._ga is None:
            raise Exception("_ga object must not be None!")
        if self.chromosome_mutation_rate_determinator is not None:
            return self.chromosome_mutation_rate_determinator
        mutator_strings = [
            ms.strip()
            for ms in self._ga.population_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        chromosome_mutation_rate_determinators: List[
            BaseChromosomeMutationRateDeterminator
        ] = []

        if definitions.RANDOM in mutator_strings:
            chromosome_mutation_rate_determinators.append(
                RandomChromosomeMutationRateDeterminator()
            )
        if definitions.CROSS_DIVERSITY in mutator_strings:
            chromosome_mutation_rate_determinators.append(
                CrossDiversityChromosomeMutationRateDeterminator()
            )
        if definitions.COST_DIVERSITY in mutator_strings:
            chromosome_mutation_rate_determinators.append(
                CostDiversityChromosomeMutationRateDeterminator()
            )
        if len(chromosome_mutation_rate_determinators) == 0:
            chromosome_mutation_rate_determinators.append(
                CostDiversityChromosomeMutationRateDeterminator()
            )
        if len(chromosome_mutation_rate_determinators) == 1:
            main_chromosome_mutation_rate_determinator: (
                BaseChromosomeMutationRateDeterminator
            ) = chromosome_mutation_rate_determinators[0]
        else:
            main_chromosome_mutation_rate_determinator = (
                ComposedChromosomeMutationRateDeterminator()
            )
            for determinator in chromosome_mutation_rate_determinators:
                main_chromosome_mutation_rate_determinator.append(determinator)
        if (
            definitions.RANDOM in mutator_strings
            and definitions.PARENT_DIVERSITY in mutator_strings
        ):
            helper_chromosome_mutation_rate_determinator: (
                BaseChromosomeMutationRateDeterminator
            ) = StrictChromosomeMutationRateDeterminator()
        else:
            helper_chromosome_mutation_rate_determinator = (
                main_chromosome_mutation_rate_determinator
            )
        return (
            main_chromosome_mutation_rate_determinator,
            helper_chromosome_mutation_rate_determinator,
        )

    def _get_chromosome_mutation_selector_combined(
        self,
    ) -> BaseChromosomeMutationSelector:
        """
        Population Mutator Instance - combined
        """
        if self._ga is None:
            raise Exception("_ga object must not be None!")
        mutator_strings = [
            ms.strip()
            for ms in self._ga.population_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        (
            main_chromosome_mutation_rate_determinator,
            helper_chromosome_mutation_rate_determinator,
        ) = self._get_chromosome_mutation_rate_determinators()
        chromosome_mutation_selectors: List[BaseChromosomeMutationSelector] = []
        if definitions.RANDOM in mutator_strings:
            chromosome_mutation_selectors.append(
                RandomChromosomeMutationSelector(
                    helper_chromosome_mutation_rate_determinator,
                    self._get_gene_mutation_selector(),
                )
            )
        if definitions.PARENT_DIVERSITY in mutator_strings:
            chromosome_mutation_selectors.append(
                ParentDiversityChromosomeMutationSelector(
                    helper_chromosome_mutation_rate_determinator,
                    self._get_gene_mutation_selector(),
                    self._get_sampling_method(
                        self._ga.parent_diversity_mutation_chromosome_sampling
                    ),
                )
            )
        if len(chromosome_mutation_selectors) == 0:
            chromosome_mutation_selectors.append(
                ParentDiversityChromosomeMutationSelector(
                    helper_chromosome_mutation_rate_determinator,
                    self._get_gene_mutation_selector(),
                    self._get_sampling_method(
                        self._ga.parent_diversity_mutation_chromosome_sampling
                    ),
                )
            )
        if len(chromosome_mutation_selectors) == 1:
            return chromosome_mutation_selectors[0]
        else:
            chromosome_mutation_selector = ComposedChromosomeMutationSelector(
                main_chromosome_mutation_rate_determinator,
                self._get_gene_mutation_selector(),
            )
            for selector in chromosome_mutation_selectors:
                chromosome_mutation_selector.append(selector)
        return chromosome_mutation_selector

    def _get_gene_mutation_rate_determinators(self):
        if self.gene_mutation_rate_determinator is not None:
            return (
                self.gene_mutation_rate_determinator,
                self.gene_mutation_rate_determinator,
            )
        mutator_strings = [
            ms.strip()
            for ms in self._ga.chromosome_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        gene_mutation_rate_determinators = []

        if definitions.RANDOM in mutator_strings:
            gene_mutation_rate_determinators.append(
                RandomGeneMutationRateDeterminator()
            )
        if definitions.CROSS_DIVERSITY in mutator_strings:
            gene_mutation_rate_determinators.append(
                CrossDiversityGeneMutationRateDeterminator()
            )
        if len(gene_mutation_rate_determinators) == 0:
            gene_mutation_rate_determinators.append(
                RandomGeneMutationRateDeterminator()
            )
            gene_mutation_rate_determinators.append(
                CrossDiversityGeneMutationRateDeterminator()
            )
        if len(gene_mutation_rate_determinators) == 1:
            main_gene_mutation_rate_determinator = gene_mutation_rate_determinators[0]
        else:
            main_gene_mutation_rate_determinator = (
                ComposedGeneMutationRateDeterminator()
            )
            for determinator in gene_mutation_rate_determinators:
                main_gene_mutation_rate_determinator.append(determinator)
        if (
            definitions.RANDOM in mutator_strings
            and definitions.COST_DIVERSITY in mutator_strings
        ):
            helper_gene_mutation_rate_determinator = (
                StrictGeneMutationRateDeterminator()
            )
        else:
            helper_gene_mutation_rate_determinator = (
                main_gene_mutation_rate_determinator
            )
        return (
            main_gene_mutation_rate_determinator,
            helper_gene_mutation_rate_determinator,
        )

    def _get_gene_mutation_selector_combined(self) -> BaseGeneMutationSelector:
        """
        Chromosome Mutator Instance - combined
        """
        if self._ga is None:
            raise Exception("_ga object must not be None!")
        mutator_strings = [
            ms.strip()
            for ms in self._ga.chromosome_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        main_gene_mutation_rate_determinator, helper_gene_mutation_rate_determinator = (
            self._get_gene_mutation_rate_determinators()
        )
        gene_mutation_selectors = []
        if definitions.RANDOM in mutator_strings:
            gene_mutation_selectors.append(
                RandomGeneMutationSelector(
                    helper_gene_mutation_rate_determinator, self.get_gene_mutator()
                ),
            )
        if definitions.CROSS_DIVERSITY in mutator_strings:
            gene_mutation_selectors.append(
                CrossDiversityGeneMutationSelector(
                    helper_gene_mutation_rate_determinator,
                    self.get_gene_mutator(),
                    self._get_sampling_method(
                        self._ga.cross_diversity_mutation_gene_sampling
                    ),
                )
            )
        if len(gene_mutation_selectors) == 0:
            gene_mutation_selectors.append(
                CrossDiversityGeneMutationSelector(
                    helper_gene_mutation_rate_determinator,
                    self.get_gene_mutator(),
                    self._get_sampling_method(
                        self._ga.cross_diversity_mutation_gene_sampling
                    ),
                )
            )
        if len(gene_mutation_selectors) == 1:
            gene_mutation_selector = gene_mutation_selectors[0]
        else:
            gene_mutation_selector = self.get_composed_gene_mutation_selector(
                main_gene_mutation_rate_determinator,
                gene_mutation_selectors,
                helper_gene_mutation_rate_determinator,
            )
        return gene_mutation_selector

    def get_composed_gene_mutation_selector(
        self,
        main_gene_mutation_rate_determinator,
        gene_mutation_selectors,
        helper_gene_mutation_rate_determinator,
    ):
        if not gene_mutation_selectors:
            gene_mutation_selectors.append(
                CrossDiversityGeneMutationSelector(
                    helper_gene_mutation_rate_determinator,
                    self.get_gene_mutator(),
                    self._get_sampling_method(
                        self._ga.cross_diversity_mutation_gene_sampling
                    ),
                )
            )
            gene_mutation_selectors.append(
                RandomGeneMutationSelector(helper_gene_mutation_rate_determinator)
            )
        gene_mutation_selector = ComposedGeneMutationSelector(
            main_gene_mutation_rate_determinator, self.get_gene_mutator()
        )
        for selector in gene_mutation_selectors:
            gene_mutation_selector.append(selector)
        return gene_mutation_selector

    def _get_parent_selector(self) -> BaseParentSelector:
        """
        Parent Selector Instance
        """
        if self._ga is None:
            raise Exception("_ga object must not be None!")
        return SamplingParentSelector(
            self._get_sampling_method(self._ga.parent_selection)
        )

    def _get_sampling_method(self, str) -> BaseSampling:
        """
        Sampling Methos Instance
        """
        str_value = str
        sampling_method_strings = str.split(definitions.PARAM_SEPARATOR)
        other_value = None
        if len(sampling_method_strings) > 1:
            str_value = sampling_method_strings[0]
            try:
                other_value = int(sampling_method_strings[1])
            except Exception:
                pass
        if str_value == definitions.TOURNAMENT:
            return TournamentSampling(other_value)
        elif str_value == definitions.FROM_TOP_TO_BOTTOM:
            return FromTopToBottomSampling()
        elif str_value == definitions.RANDOM:
            return RandomSampling()
        return RouletteWheelSampling()

    def _get_exit_checker(self) -> BaseExitChecker:
        """
        Exit Checker Instance
        """
        if self._ga is None:
            raise Exception("ga object must not be None!")
        if self._ga.exit_check == definitions.AVG_COST:
            return AvgCostExitChecker(self._ga.max_attempt_no)
        if self._ga.exit_check == definitions.MIN_COST:
            return MinCostExitChecker(self._ga.max_attempt_no)
        if self._ga.exit_check == definitions.GENERATIONS:
            return NumberOfGenerationsExitChecker(self._ga.number_of_generations)
        return RequestedCostExitChecker(self._ga.requested_cost)

    def _get_crossover(self) -> BaseCrossover:
        """
        Crossover Instance
        """
        if self._ga is None:
            raise Exception("ga object must not be None!")
        mutator_strings = [
            ms.strip()
            for ms in self._ga.population_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        population_mutation_selection_strings = [
            value
            for value in mutator_strings
            if value in definitions.POPULATION_MUTATION_SELECTION_STRINGS
        ]
        if (
            not population_mutation_selection_strings
            or definitions.PARENT_DIVERSITY in mutator_strings
        ):
            return BlendingParentDiversityCrossover(
                self.get_gene_mutation_selector(),
                self.get_chromosome_immigrator(),
            )
        if self._ga.crossover == definitions.BLENDING:
            return BlendingCrossover(
                self.get_gene_mutation_selector(),
                self.get_chromosome_immigrator(),
            )
        if self._ga.crossover == definitions.UNIFORM:
            return UniformCrossover(
                self.get_gene_mutation_selector(),
                self.get_chromosome_immigrator(),
            )
        return BlendingParentDiversityCrossover(
            self.get_gene_mutation_selector(),
            self.get_chromosome_immigrator(),
        )

    def _get_variable_updater(self):
        """
        Variable Updater Instance
        """
        return CommonVariableUpdater()

    def _get_population_updater(self):
        """
        Population Updater Instance
        """
        return CommonPopulationUpdater()



import sys
import traceback
from abc import ABC, abstractmethod
from typing import Optional

from gadapt.ga_model.population import Population
from gadapt.ga_model.chromosome import Chromosome


class BaseCostFinder(ABC):
    """
    Base class for cost finding
    """

    def __init__(self):
        super().__init__()
        self.population: Optional[Population] = None

    def _execute_function(self, cost_function, c: Chromosome):
        """
        Executes the cost function

        Args:
            cost_function: Function to execute
            c (Chromosome): The chromosome with
            genes containing values for the function execution.
        """
        dict = {}
        for g in c:
            dict[g.decision_variable.variable_id] = g.variable_value
        try:
            cost_value = cost_function(dict)
            c.cost_value = cost_value
        except Exception:
            print(Exception)
            traceback.print_exc()
            c.succ = False
            c.cost_value = sys.float_info.max

    @abstractmethod
    def _find_costs_for_population(self):
        pass

    def find_costs(self, population):
        """
        Finds costs for the population

        Args:
            population (Population): The population to find costs for each chromosome
        """
        self.population = population
        self._find_costs_for_population()
        self.population.min_cost_per_generation.append(self.population.min_cost)



import random

from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)
from gadapt.operations.mutation.chromosome_mutation.random_gene_mutation_selector import (
    RandomGeneMutationSelector,
)
from gadapt.operations.sampling.base_sampling import BaseSampling
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class CrossDiversityGeneMutationSelector(RandomGeneMutationSelector):
    """
    Selects and mutates a chromosome based on the cross diversity of decision variables in the population.
    """

    def __init__(
        self,
        gene_mutation_rate_determinator: BaseGeneMutationRateDeterminator,
        gene_mutator: BaseGeneMutator,
        sampling: BaseSampling,
    ) -> None:
        super().__init__(gene_mutation_rate_determinator, gene_mutator)
        self._sampling = sampling

    def _mutate_chromosome(self):
        if self.number_of_mutation_genes == 0:
            self.number_of_mutation_genes = 1
        x_genes = [g for g in self.chromosome]
        x_genes.sort(key=lambda g: -g.decision_variable.cross_diversity_coefficient)
        number_of_mutation_genes = (
            self._gene_mutation_rate_determinator.get_number_of_mutation_genes(
                self.chromosome, self.number_of_mutation_genes
            )
        )
        if number_of_mutation_genes > len(x_genes):
            number_of_mutation_genes = len(x_genes)
        if number_of_mutation_genes == 0:
            max_number_of_mutation_genes = 1
        else:
            max_number_of_mutation_genes = random.randint(1, number_of_mutation_genes)
        genes_for_mutation = self._sampling.get_sample(
            x_genes,
            max_number_of_mutation_genes,
            lambda g: g.decision_variable.cross_diversity_coefficient,
        )
        for g in genes_for_mutation:
            self._gene_mutator.mutate(g)
        return len(genes_for_mutation)



from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.sampling.base_sampling import BaseSampling


class SamplingParentSelector(BaseParentSelector):
    """
    Parent Selector based on sampling.
    Sampling is the algorithm for extracting a sample from the population,
    based on specific value of the chromosme.
    In this case sampling depends on cost value.

    Selects mates for mating from the population
    """

    def __init__(self, sampling: BaseSampling) -> None:
        super().__init__()
        self._sampling = sampling

    def _select_potential_parents(self):
        return self._sampling.get_sample(
            self.population.chromosomes, len(self.population), lambda c: c.cost_value
        )



import random

from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class BlendingCrossover(BaseCrossover):
    """
    Blending Crossover combines
    gene values from the two parents into new variable values in offsprings.
    One value of the offspring variable comes from a combination of the two
    corresponding values of the parental genes
    """

    def __init__(
        self,
        mutator: BaseGeneMutationSelector,
        immigrator: BaseChromosomeImmigrator,
    ):
        super(BlendingCrossover, self).__init__(mutator, immigrator)
        self._current_gene_number = -1

    def _combine(self):
        decision_variable = self._father_gene.decision_variable
        val_father = self._father_gene.variable_value
        val_mother = self._mother_gene.variable_value
        x = 1
        if val_mother > val_father:
            x = -1
        beta_steps = random.randint(
            0, round(abs((val_father - val_mother) / decision_variable.step))
        )
        val1 = round(
            val_father - (beta_steps * x) * decision_variable.step,
            decision_variable.decimal_places,
        )
        val2 = round(
            val_mother + (beta_steps * x) * decision_variable.step,
            decision_variable.decimal_places,
        )
        return val1, val2



import random
from typing import Tuple

from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)


class UniformCrossover(BaseCrossover):
    """
    Uniform Crossover. Genes from parents' chromosomes are combined in a uniform way.
    """

    def __init__(
        self,
        mutator: BaseGeneMutationSelector,
        immigrator: BaseChromosomeImmigrator,
    ):
        super(UniformCrossover, self).__init__(mutator, immigrator)

    def _combine(self) -> Tuple[float, float]:
        rnd = random.randint(0, 2)
        if rnd == 0:
            return self._father_gene.variable_value, self._mother_gene.variable_value
        elif rnd == 1:
            return self._father_gene.variable_value, self._father_gene.variable_value
        else:
            return self._mother_gene.variable_value, self._mother_gene.variable_value



"""
Genetic algorithm options
"""

from typing import List

from gadapt.ga_model.decision_variable import DecisionVariable


class GAOptions:
    def __init__(self, ga) -> None:
        """
        Genetic algorithm options class
        Args:
            ga: Genetic algorithm class, containing data to initialize options
        """
        super().__init__()
        self._population_size = ga.population_size
        self._keep_elitism_percentage = ga.keep_elitism_percentage
        self._number_of_crossover_parents = ga.number_of_crossover_parents
        self._cost_function = ga.cost_function
        self._immigration_number = ga.immigration_number
        self._set_number_of_mutation_chromosomes(ga)
        self._max_attempt_no = ga.max_attempt_no
        self._requested_cost = ga.requested_cost
        self._logging = ga.logging
        self._decision_variables = ga._decision_variables
        self._set_number_of_mutation_genes(ga)
        self._must_mutate_for_same_parents = ga.must_mutate_for_same_parents
        self._timeout = ga.timeout

    def _set_number_of_mutation_chromosomes(self, ga):
        if (
            (ga.number_of_mutation_chromosomes is not None)
            and isinstance(ga.number_of_mutation_chromosomes, int)
            and ga.number_of_mutation_chromosomes >= 0
        ):
            self._number_of_mutation_chromosomes = ga.number_of_mutation_chromosomes
        elif (
            (ga.percentage_of_mutation_chromosomes is not None)
            and isinstance(ga.percentage_of_mutation_chromosomes, float)
            and 0.0 <= ga.percentage_of_mutation_chromosomes <= 100
        ):
            nomc = round(
                (self.population_size / 2)
                * (ga.percentage_of_mutation_chromosomes / 100)
            )
            while (
                nomc
                >= round(self.population_size * (self.keep_elitism_percentage / 100))
                - self.immigration_number
            ):
                nomc -= 1
            if nomc < 0:
                raise Exception(
                    "Invalid percentage of mutation chromosomes and immigration number"
                )
            self._number_of_mutation_chromosomes = nomc
        else:
            self._number_of_mutation_chromosomes = ga.number_of_mutation_chromosomes

    def _set_number_of_mutation_genes(self, ga):
        if (
            (ga.number_of_mutation_genes is not None)
            and isinstance(ga.number_of_mutation_genes, int)
            and ga.number_of_mutation_genes >= 0
        ):
            self._number_of_mutation_genes = ga.number_of_mutation_genes
        elif (
            (ga.percentage_of_mutation_genes is not None)
            and isinstance(ga.percentage_of_mutation_genes, float)
            and 0.0 <= ga.percentage_of_mutation_genes <= 100
        ):
            self._number_of_mutation_genes = round(
                float(len(self._decision_variables))
                * (ga.percentage_of_mutation_genes / 100)
            )
        else:
            self._number_of_mutation_genes = ga.number_of_mutation_genes

    @property
    def requested_cost(self) -> float:
        """
        Determines the requested value which causes the exit from the genetic algorithm
        """
        return self._requested_cost

    @requested_cost.setter
    def requested_cost(self, value: float):
        self._requested_cost = value

    @property
    def max_attempt_no(self) -> int:
        """
        Determines the number of generations in which there is no\
            improvement in the average/minimal cost.
        """
        return self._max_attempt_no

    @max_attempt_no.setter
    def max_attempt_no(self, value: int):
        self._max_attempt_no = value

    @property
    def immigration_number(self) -> int:
        """
        Number of immigration chromosomes
        """
        return self._immigration_number

    @immigration_number.setter
    def immigration_number(self, value: int):
        self._immigration_number = value

    @property
    def number_of_mutation_chromosomes(self) -> int:
        """
        The number of mutation chromosomes in the population.
        """
        return self._number_of_mutation_chromosomes

    @number_of_mutation_chromosomes.setter
    def number_of_mutation_chromosomes(self, value: int):
        self._number_of_mutation_chromosomes = value

    @property
    def number_of_mutation_genes(self) -> int:
        """
        The number of mutated genes in each chromosome.
        """
        return self._number_of_mutation_genes

    @number_of_mutation_genes.setter
    def number_of_mutation_genes(self, value: int):
        self._number_of_mutation_genes = value

    @property
    def cost_function(self):
        """
        Custom function for the cost calculation (fitness)
        """
        return self._cost_function

    @cost_function.setter
    def cost_function(self, value):
        self._cost_function = value

    @property
    def population_size(self) -> int:
        """
        Number of chromosomes in the population.
        """
        return self._population_size

    @population_size.setter
    def population_size(self, value: int):
        self._population_size = value

    @property
    def keep_elitism_percentage(self) -> int:
        """
        Percentage number of chromosomes to be kept in the population by the cost value
        """
        return self._keep_elitism_percentage

    @keep_elitism_percentage.setter
    def keep_elitism_percentage(self, value: int):
        self._keep_elitism_percentage = value

    @property
    def number_of_crossover_parents(self) -> int:
        """
        Number of parents to be included in the mating pool
        """
        if self._number_of_crossover_parents > 1:
            return self._number_of_crossover_parents
        return self.keep_number

    @number_of_crossover_parents.setter
    def number_of_crossover_parents(self, value: int):
        self._number_of_crossover_parents = value

    @property
    def decision_variables(self) -> List[DecisionVariable]:
        """
        Collection of decision variables
        """
        return self._decision_variables

    @property
    def _abandon_number(self) -> int:
        return self._get_abandon_number()

    def _get_abandon_number(self) -> int:
        return round(self.population_size * (1 - self.keep_elitism_percentage / 100))

    @property
    def keep_number(self) -> int:
        return self.population_size - self._abandon_number

    @property
    def logging(self) -> bool:
        """
        If True, the log file will be created in the\
            current working directory.
        """
        return self._logging

    @logging.setter
    def logging(self, value: bool):
        self._logging = value

    @property
    def must_mutate_for_same_parents(self) -> bool:
        """
        Indicates if completely the same parents must influence\
            mutation for their children.
        """
        return self._must_mutate_for_same_parents

    @must_mutate_for_same_parents.setter
    def must_mutate_for_same_parents(self, value: bool):
        self._must_mutate_for_same_parents = value

    @property
    def timeout(self) -> int:
        """
        A number of seconds after which the genetic algorithm optimisation will\
            exit, regardless of whether exit_check criteria is reached.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value: int):
        self._timeout = value



from abc import ABC, abstractmethod
from typing import Optional

from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.cost_finding.base_cost_finder import BaseCostFinder
from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.immigration.population_immigration.base_population_immigrator import (
    BasePopulationImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_selector import (
    BaseGeneMutationSelector,
)
from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator
from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_selector import (
    BaseChromosomeMutationSelector,
)

"""
    Factory definition for creating  class instances based on GA options
"""


class BaseGAFactory(ABC):

    def __init__(self) -> None:
        super().__init__()
        self.population_updater = None
        self._ga = None
        self._options = None
        self.variable_updater = None
        self.cost_finder: Optional[BaseCostFinder] = None
        self.population_immigrator: Optional[BasePopulationImmigrator] = None
        self.chromosome_immigrator: Optional[BaseChromosomeImmigrator] = None
        self.gene_mutator: Optional[BaseGeneMutator] = None
        self.gene_mutation_selector: Optional[BaseGeneMutationSelector] = None
        self.chromosome_mutation_selector: Optional[BaseChromosomeMutationSelector] = (
            None
        )
        self.parent_selector: Optional[BaseParentSelector] = None
        self.exit_checker: Optional[BaseExitChecker] = None
        self.crossover: Optional[BaseCrossover] = None

    def initialize_factory(self, ga):
        self._ga = ga
        self._options = GAOptions(ga)

    def get_cost_finder(self) -> BaseCostFinder:
        """
        Cost Finder instance
        """
        if self.cost_finder is None:
            self.cost_finder = self._get_cost_finder()
        return self.cost_finder

    def get_population_immigrator(self) -> BasePopulationImmigrator:
        """
        Population Immigrator Instance
        """
        if self.population_immigrator is None:
            self.population_immigrator = self._get_population_immigrator()
        return self.population_immigrator

    def get_chromosome_immigrator(self) -> BaseChromosomeImmigrator:
        """
        Chromosome Immigrator Instance
        """
        if self.chromosome_immigrator is None:
            self.chromosome_immigrator = self._get_chromosome_immigrator()
        return self.chromosome_immigrator

    def get_gene_mutation_selector(self) -> BaseGeneMutationSelector:
        """
        Gene Mutation Selector Instance
        """
        if self.gene_mutation_selector is None:
            self.gene_mutation_selector = self._get_gene_mutation_selector()
        return self.gene_mutation_selector

    def get_gene_mutator(self) -> BaseGeneMutator:
        """
        Gene Mutator Instance
        """
        if self.gene_mutator is None:
            self.gene_mutator = self._get_gene_mutator()
        return self.gene_mutator

    def get_chromosome_mutation_selector(self) -> BaseChromosomeMutationSelector:
        """
        Chromosome Mutation Selector Instance
        """
        if self.chromosome_mutation_selector is None:
            self.chromosome_mutation_selector = self._get_chromosome_mutation_selector()
        return self.chromosome_mutation_selector

    def get_parent_selector(self) -> BaseParentSelector:
        """
        Parent Selector Instance
        """
        if self.parent_selector is None:
            self.parent_selector = self._get_parent_selector()
        return self.parent_selector

    def get_exit_checker(self) -> BaseExitChecker:
        """
        Exit Checker Instance
        """
        if self.exit_checker is None:
            self.exit_checker = self._get_exit_checker()
        return self.exit_checker

    def get_variable_updater(self):
        """
        Variable Updater Instance
        """
        if self.variable_updater is None:
            self.variable_updater = self._get_variable_updater()
        return self.variable_updater

    def get_population_updater(self):
        """
        Population Updater Instance
        """
        if self.population_updater is None:
            self.population_updater = self._get_population_updater()
        return self.population_updater

    def get_crossover(self) -> BaseCrossover:
        """
        Crossover Instance
        """
        if self.crossover is None:
            self.crossover = self._get_crossover()
        return self.crossover

    @abstractmethod
    def _get_cost_finder(self) -> BaseCostFinder:
        """
        Cost Finder instance
        """
        pass

    @abstractmethod
    def _get_population_immigrator(self) -> BasePopulationImmigrator:
        """
        Population Immigrator Instance
        """
        pass

    @abstractmethod
    def _get_chromosome_immigrator(self) -> BaseChromosomeImmigrator:
        """
        Chromosome Immigrator Instance
        """
        pass

    @abstractmethod
    def _get_gene_mutation_selector(self) -> BaseGeneMutationSelector:
        """
        Chromosome Mutator Instance
        """
        pass

    @abstractmethod
    def _get_gene_mutator(self) -> BaseGeneMutator:
        """
        Gene Mutator Instance
        """
        pass

    @abstractmethod
    def _get_chromosome_mutation_selector(self) -> BaseChromosomeMutationSelector:
        """
        Population Mutator Instance
        """
        pass

    @abstractmethod
    def _get_parent_selector(self) -> BaseParentSelector:
        """
        Parent Selector Instance
        """
        pass

    @abstractmethod
    def _get_exit_checker(self) -> BaseExitChecker:
        """
        Exit Checker Instance
        """
        pass

    @abstractmethod
    def _get_variable_updater(self):
        """
        Variable Updater Instance
        """
        pass

    @abstractmethod
    def _get_population_updater(self):
        """
        Population Updater Instance
        """
        pass

    @abstractmethod
    def _get_crossover(self) -> BaseCrossover:
        """
        Crossover Instance
        """
        pass



import math

from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class CostDiversityChromosomeMutationRateDeterminator(
    BaseChromosomeMutationRateDeterminator
):
    """
    Determines the number of chromosomes to be mutated in a population based on the cost diversity of the population.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_number_of_mutation_chromosomes(self) -> int:
        def get_mutation_rate() -> float:
            if (
                self.population.average_cost_step_in_first_population is None
                or math.isnan(self.population.average_cost_step_in_first_population)
            ):
                return 1.0
            cost_step_ratio = float(
                self.population.average_cost_step
                / self.population.average_cost_step_in_first_population
            )
            if cost_step_ratio > 1.0:
                cost_step_ratio = 1.0
            return 1.0 - cost_step_ratio

        mutation_rate = get_mutation_rate()
        f_return_value = mutation_rate * float(self.max_number_of_mutation_chromosomes)
        return round(f_return_value)



"""
Gene
"""

import math

import gadapt.adapters.string_operation.ga_strings as ga_strings
import gadapt.ga_model.definitions as definitions
from gadapt.ga_model.decision_variable import DecisionVariable
from gadapt.ga_model.ranking_model import RankingModel


class Gene(RankingModel):
    def __init__(self, gen_variable, var_value=None):
        """
        Gene class. Gene is a part of chromosome.
        It contains concrete values for decision variables.
        Args:
            gen_variable: Decision variable which defines the gene
            var_value: Value of the gene
        """
        super().__init__()
        self.decision_variable = gen_variable
        self.variable_value = var_value
        self._rank = -1
        self._cummulative_probability = definitions.FLOAT_NAN
        if self.variable_value is None or math.isnan(self.variable_value):
            self.set_random_value()

    def __str__(self) -> str:
        return self._to_string()

    def _to_string(self):
        return ga_strings.gene_to_string(self)

    @property
    def decision_variable(self) -> DecisionVariable:
        """
        Decision variable which defines the gene
        """
        return self._decision_variable

    @decision_variable.setter
    def decision_variable(self, value: DecisionVariable):
        if not isinstance(value, DecisionVariable):
            raise
        self._decision_variable = value

    @property
    def variable_value(self):
        """
        Value of the gene
        """
        return self._variable_value

    @variable_value.setter
    def variable_value(self, value):
        self._variable_value = value

    def set_random_value(self):
        """
        Sets a random value for the variable_value property
        """
        self.variable_value = self.decision_variable.make_random_value()



"""
The main genetic algorithm module
"""

import sys
from typing import List, Optional

import gadapt.ga_model.definitions as definitions
import gadapt.utils.ga_utils as ga_utils
from gadapt.adapters.validation.common_options_validator import CommonOptionsValidator
from gadapt.execution.ga_executor import GAExecutor
from gadapt.factory.ga_base_factory import BaseGAFactory
from gadapt.factory.ga_factory import GAFactory
from gadapt.ga_model.decision_variable import DecisionVariable
from gadapt.ga_model.ga_options import GAOptions
from gadapt.ga_model.ga_results import GAResults
import copy


class GA:
    """
    The main genetic algorithm class
    """

    def __init__(
        self,
        cost_function=None,
        population_size=32,
        keep_elitism_percentage=50,
        number_of_crossover_parents=-1,
        exit_check=definitions.AVG_COST,
        requested_cost=sys.float_info.max,
        number_of_generations=200,
        max_attempt_no=2,
        parent_selection=definitions.ROULETTE_WHEEL,
        crossover=definitions.BLENDING,
        population_mutation="{0}{1}{2}{3}{4}".format(
            definitions.COST_DIVERSITY,
            definitions.PARAM_SEPARATOR,
            definitions.CROSS_DIVERSITY,
            definitions.PARAM_SEPARATOR,
            definitions.PARENT_DIVERSITY,
        ),
        number_of_mutation_chromosomes=-1,
        percentage_of_mutation_chromosomes=50.0,
        parent_diversity_mutation_chromosome_sampling=definitions.ROULETTE_WHEEL,
        must_mutate_for_same_parents=True,
        chromosome_mutation=f"{definitions.CROSS_DIVERSITY}{definitions.PARAM_SEPARATOR}{definitions.RANDOM}",
        gene_mutation=f"{definitions.CROSS_DIVERSITY}{definitions.PARAM_SEPARATOR}{definitions.RANDOM}",
        number_of_mutation_genes=-1,
        percentage_of_mutation_genes=20.0,
        cross_diversity_mutation_gene_sampling=definitions.ROULETTE_WHEEL,
        immigration_number=0,
        logging=False,
        timeout=120,
        factory: Optional[BaseGAFactory] = None,
    ) -> None:
        """
        The constructor of the GA class accepts all parameters required to
        create an instance of the GA class. It validates such parameters.

        Args:
            cost_function: Custom function for the cost calculation (fitness).
            The optimisation
            goal is minimising the output of the cost function.
                cost_function must be the function with one argument - a dictionary of
                values, where the key is an index (the ordinal of adding parameters)
                and the key is the parameter's value to be optimised.

                When adding parameters, there should be as
                many parameters as the function uses. The
                cost_function is the only mandatory parameter.

            population_size: Number of chromosomes in the population.

            exit_check: A criteria for the exit for the genetic algorithm

            requested_cost: This parameter only takes place when exit_check
            has value “requested”. It determines the requested value
            which causes the exit from the genetic algorithm

            number_of_generations: This parameter only takes place when exit_check
            has value “generations”. It determines the number of generations
            after which the genetic algorithm exits

            max_attempt_no: This parameter only takes place when exit_check
            has value “avg_cost” or “min_cost”. It determines the number of
            generations in which there is no improvement in the average/minimal cost.

            parent_selection: The sampling algorithm in parent selection.

            population_mutation: A type of mutation for the entire population.

                Based on the value of this parameter, the number of mutation
                chromosomes can be determined, along with how
                chromosomes for the mutation will be selected.

            number_of_mutation_chromosomes: The number of
            mutation chromosomes in the population.

                In case it's value is equal to or higher than 0,
                it overrides percentage_of_mutation_chromosomes.

                This value is the upper bound - the actual
                number of mutated chromosomes
                can vary from 1 to number_of_mutation_chromosomes.

            percentage_of_mutation_chromosomes: The percentage of mutated chromosomes
            in the population.
                This value is applied to the population_size value and rounded to an
                integer value, giving the number of mutation chromosomes.

                For example, if population_size has a value of 32,
                and percentage_of_mutation_chromosomes has a value of
                10, the number of mutation chromosomes will be 3.

                The calculated value is an upper bound - the actual
                number of mutated chromosomes can vary from 1 to the calculated value.

                percentage_of_mutation_chromosomes only applies if
                number_of_mutation_chromosomes does not have a valid
                integer value equal to or higher than 0.

            parent_diversity_mutation_chromosome_sampling: The
            sampling algorithm for mutating chromosomes when population_mutation
            contains value “parent_diversity”.
                It only applies when population_mutation has value
                “parent_diversity”. It determines the way how chromosomes are to
                be selected based on the diversity of their parents.

            must_mutate_for_same_parents: Indicates if completely the same
            parents must influence mutation for their children.
                In other words, each child will be mutated if it has
                parents with a diversity value of 0.

                If must_mutate_for_same_parents has the value True, the
                number of mutated chromosomes can outreach value
                determined by number_of_mutation_chromosomes or
                percentage_of_mutation_chromosomes

            chromosome_mutation: The type of mutation of genes in chromosomes.

            gene_mutation: The type of assigning values to genes in the mutation

            number_of_mutation_genes: The number of mutated genes in each chromosome.
                In case it's value is equal to or higher than 0, it overrides
                percentage_of_mutation_genes.

                This value is the upper bound - the number of mutated genes
                can vary from 1 to number_of_mutation_genes.

            percentage_of_mutation_genes: The percentage of
            mutated genes in each chromosome.
                It applies to the chromosome size
                (number of genes in each chromosome), and
                the calculated value rounds to an integer value.

                The calculated value is the upper bound - the actual number of
                mutated genes can vary from 1 to the calculated value.

                percentage_of_mutation_genes only applies if
                number_of_mutations_genes does not have a valid
                integer value equal to or higher than 0.

            cross_diversity_mutation_gene_sampling: The sampling algorithm for
            mutating chromosomes when chromosome_mutation has value “cross_diversity”.
                It only applies when chromosome_mutation has value
                “cross_diversity” . It determines the way how genes
                are to be selected based on the cross-diversity.

            immigration_number: Refers to the “Random Immigrants”
            concepts. This strategy introduces a certain number of
            individuals into the population during the evolution process.
                These new individuals are generated randomly
                and injected into the population.

            logging: If this parameter has a True value, the log file
            will be created in the current working directory.
            The log file contains the flow of genetic algorithm execution,
            along with values of chromosomes, genes and cost
            functions in each generation

            timeout: A number of seconds after which the genetic algorithm
            optimisation will exit, regardless of whether
            exit_check criteria is reached.

            factory: Factory for creating object
        """
        self.cost_function = cost_function
        self.population_size = population_size
        self.keep_elitism_percentage = keep_elitism_percentage
        self.number_of_crossover_parents = number_of_crossover_parents
        self.exit_check = exit_check
        self.requested_cost = requested_cost
        self.number_of_generations = number_of_generations
        self.max_attempt_no = max_attempt_no
        self.parent_selection = parent_selection
        self.crossover = crossover
        self.population_mutation = population_mutation
        self.must_mutate_for_same_parents = must_mutate_for_same_parents
        self.number_of_mutation_chromosomes = number_of_mutation_chromosomes
        self.percentage_of_mutation_chromosomes = percentage_of_mutation_chromosomes
        self.number_of_mutation_genes = number_of_mutation_genes
        self.percentage_of_mutation_genes = percentage_of_mutation_genes
        self.chromosome_mutation = chromosome_mutation
        self.gene_mutation = gene_mutation
        self.immigration_number = immigration_number
        self.logging = logging
        self._decision_variables: List[DecisionVariable] = []
        self.cross_diversity_mutation_gene_sampling = (
            cross_diversity_mutation_gene_sampling
        )
        self.parent_diversity_mutation_chromosome_sampling = (
            parent_diversity_mutation_chromosome_sampling
        )
        self.timeout = timeout
        self._current_dv_id = 0
        self._factory = factory

    def execute(self) -> GAResults:
        """
        Executes genetic algorithm optimization base on the
        provided parameters and arguments.
        """
        validator = CommonOptionsValidator(self)
        validator.validate()
        if not validator.success:
            results = GAResults()
            results.success = False
            results.messages = validator.validation_messages
            return results
        ga_options = GAOptions(self)
        factory: BaseGAFactory = self.get_factory()
        factory.initialize_factory(self.clone())
        return GAExecutor(ga_options, factory).execute()

    def get_factory(self) -> BaseGAFactory:
        if self._factory:
            return self._factory
        return GAFactory()

    @property
    def population_size(self) -> int:
        """
        Population size number
        """
        return self._population_size

    @population_size.setter
    def population_size(self, value: int):
        self._population_size = ga_utils.try_get_int(value)

    @property
    def keep_elitism_percentage(self) -> int:
        """
        Percentage number of chromosomes to be kept in the population by the cost value
        """
        return self._keep_elitism_percentage

    @keep_elitism_percentage.setter
    def keep_elitism_percentage(self, value: int):
        self._keep_elitism_percentage = ga_utils.try_get_int(value)

    @property
    def number_of_crossover_parents(self) -> int:
        """
        Number of parents to be included in the mating pool
        """
        return self._number_of_crossover_parents

    @number_of_crossover_parents.setter
    def number_of_crossover_parents(self, value: int):
        self._number_of_crossover_parents = ga_utils.try_get_int(value)

    def add(
        self,
        min_value: float = -sys.float_info.max,
        max_value: float = sys.float_info.max,
        step: float = sys.float_info.min,
    ):
        """
        Adds variables to be optimized.
        Args:
            min_value (float): lower bound of possible variable values
            max_value (float): upper bound of possible variable values
            step (float): the step that will be used in changing the
            variables values during the optimization
        """
        if step < 0.000000000000001:
            step = 0.000000000000001
        if (not isinstance(min_value, float) and not isinstance(min_value, int)) or (
            not isinstance(max_value, float) and not isinstance(max_value, int)
        ):
            raise Exception("min value, max value and step must be numerical values!")
        decision_variable = DecisionVariable(self._current_dv_id)
        decision_variable.min_value = min_value
        decision_variable.max_value = max_value
        decision_variable.step = step
        self._decision_variables.append(decision_variable)
        self._current_dv_id += 1

    @property
    def cost_function(self):
        """
        Custom function for the cost calculation (fitness). The optimisation goal
        is minimising the output of the cost function.
        cost_function must be the function with one argument -
        a dictionary of values, where the key is an index
        (the ordinal of adding parameters)
        and the key is the parameter's value to be optimised.
        When adding parameters, there should be as many parameters
        as the function uses. The cost_function is the
        only mandatory parameter.
        """
        return self._cost_function

    @cost_function.setter
    def cost_function(self, value):
        self._cost_function = value

    @property
    def number_of_mutation_genes(self) -> int:
        """
        The number of mutated genes in each chromosome.
        In case it's value is equal to or higher than 0, it overrides
        percentage_of_mutation_genes.
        This value is the upper bound - the number of mutated genes
        can vary from 1 to number_of_mutation_genes.
        """
        return self._number_of_mutation_genes

    @number_of_mutation_genes.setter
    def number_of_mutation_genes(self, value: int):
        self._number_of_mutation_genes = ga_utils.try_get_int(value)

    @property
    def percentage_of_mutation_genes(self) -> float:
        """
        The percentage of mutated genes in each chromosome.
        It applies to the chromosome size (number of genes in each chromosome),
        and the calculated value rounds to an integer value.
        The calculated value is the upper bound - the actual number of mutated
        genes can vary from 1 to the calculated value.
        percentage_of_mutation_genes only applies if number_of_mutations_genes
        does not have a valid integer value equal to or higher than 0.
        """
        return self._percentage_of_mutation_genes

    @percentage_of_mutation_genes.setter
    def percentage_of_mutation_genes(self, value: float):
        self._percentage_of_mutation_genes = ga_utils.try_get_float(value)

    @property
    def number_of_mutation_chromosomes(self) -> int:
        """
        The number of mutation chromosomes in the population.
        In case it's value is equal to or higher than 0, it
        overrides percentage_of_mutation_chromosomes.
        This value is the upper bound - the actual number of
        mutated chromosomes can vary from 1 to number_of_mutation_chromosomes.
        """
        return self._number_of_mutation_chromosomes

    @number_of_mutation_chromosomes.setter
    def number_of_mutation_chromosomes(self, value: int):
        self._number_of_mutation_chromosomes = ga_utils.try_get_int(value)

    @property
    def percentage_of_mutation_chromosomes(self) -> float:
        """
        The percentage of mutated chromosomes in the population.

        This value is applied to the population_size value and
        rounded to an integer value, giving the number
        of mutation chromosomes.

        For example, if population_size has a value of 32, and
        percentage_of_mutation_chromosomes has a value of 10,
        the number of mutation chromosomes will be 3.

        The calculated value is an upper bound - the actual number of
        mutated chromosomes can vary from 1 to the calculated value.
        percentage_of_mutation_chromosomes only applies
        if number_of_mutation_chromosomes
        does not have a valid integer value equal to or higher than 0.
        """
        return self._percentage_of_mutation_chromosomes

    @percentage_of_mutation_chromosomes.setter
    def percentage_of_mutation_chromosomes(self, value: float):
        self._percentage_of_mutation_chromosomes = ga_utils.try_get_float(value)

    @property
    def immigration_number(self) -> int:
        """
        Refers to the “Random Immigrants” concepts. This strategy introduces
        a certain number of individuals into the population
        during the evolution process.
        """
        return self._immigration_number

    @immigration_number.setter
    def immigration_number(self, value: int):
        self._immigration_number = ga_utils.try_get_int(value)

    @property
    def population_mutation(self) -> str:
        """
        A type of mutation for the entire population.

        Supported values:

        **"cost_diversity"** - It applies to the number of mutation chromosomes.
        “cost_diversity” determines the number of mutated chromosomes adaptively,
        using the diversity of costs in the population. Lower cost diversity
        means a higher number of mutated chromosomes. The minimal value of
        mutated chromosomes is 0, and the maximal value is determined by
        the value of number_of_mutation_chromosomes or
        percentage_of_mutation_chromosomes
        parameters. If population_mutation has a value other than
        “cost_diversity”, the number of mutation chromosomes is a
        random value from 1 to number_of_mutation_chromosomes
        value (or to value determined by percentage_of_mutation_chromosomes value).
        “cost_diversity” means that the “parent_diversity” method is selected
        to select chromosomes to be mutated. This method only determines
        the number of mutated chromosomes, but not how
        chromosomes are selected for the mutation.

        **"parent_diversity"** - It applies to the way how mutation
        chromosomes will be selected. “parent_diversity” selects
        chromosomes to be mutated using the diversity of their parents.
        The more similar parents (lower parent diversity) mean a higher
        probability of mutation for the child. Based on the calculated
        parent diversity, chromosomes may be selected by one of the
        sampling methods, which is determined by the value of the
        parent_diversity_mutation_chromosome_sampling parameter.

        **"random"** - It applies to the number of mutation chromosomes
        and to the way how mutation chromosomes will be selected. “random”
        selects chromosomes to be mutated randomly, and randomly
        determines the number of mutated chromosomes (with the
        upper bound of number_of_mutation_chromosomes)
        """
        return self._population_mutation

    @population_mutation.setter
    def population_mutation(self, value: str):
        self._population_mutation = ga_utils.prepare_string(value)

    @property
    def parent_diversity_mutation_chromosome_sampling(self) -> str:
        """
        The sampling algorithm for mutating chromosomes when
        population_mutation contains value “parent_diversity”.
        It only applies when population_mutation has value
        “parent_diversity”. It determines the way how chromosomes are
        to be selected based on the diversity of their parents.

        Supported values:

        **"roulette_wheel"** - The Roulette Wheel sampling algorithm
        (also known as “Weighted Random Pairing”). The probabilities
        assigned to the chromosomes to be mutated are proportional
        to the similarity of their parents (inversely proportional
        to the parent diversity). A chromosome with the lowest
        parent diversity has the greatest probability of
        mutation, while the chromosome with the highest
        parent diversity has the lowest probability of mutation.

        **"tournament"** - The Tournament sampling algorithm.
        It randomly picks small subsets (groups) of chromosomes,
        and chromosomes with the lowest parent diversity (highest parent similarity)
        in subsets are chosen to be mutated. “tournament” can have an additional
        parameter separated from the “tournament” keyword by the comma.
        The other value represents a group size. For example, “tournament,8” means
        that the tournament mutation sampling algorithm is chosen, and
        each group contains up to 8 members. The default group size is 4.

        **"from_top_to_bottom"** - From Top To Bottom sampling
        algorithm starts at the top of the list and
        selects chromosomes for mutation.

        **"random"** - Random selection algorithm uses a
        uniform random number generator to select
        chromosomes for mutation. In this case,
        sampling for mutation will not depend
        on parent diversity.
        """
        return self._parent_diversity_mutation_chromosome_sampling

    @parent_diversity_mutation_chromosome_sampling.setter
    def parent_diversity_mutation_chromosome_sampling(self, value: str):
        self._parent_diversity_mutation_chromosome_sampling = ga_utils.prepare_string(
            value
        )

    @property
    def chromosome_mutation(self) -> str:
        """
        The type of mutation of genes in chromosomes.

        Supported values:

        **"cross_diversity"** - Considers the diversity of genes of the same
        type in the population. Lower diversity can mean
        that this decision variable approaches some local
        minimums, and therefore such genes increase the
        chance for mutation. Based on the calculated
        cross-diversity, chromosomes may be selected by
        one of the sampling methods, which is
        determined by the value of the
        cross_diversity_mutation_gene_sampling parameter.

        **"random"** - Genes are randomly selected for the mutation
        """
        return self._chromosome_mutation

    @chromosome_mutation.setter
    def chromosome_mutation(self, value: str):
        self._chromosome_mutation = ga_utils.prepare_string(value)

    @property
    def gene_mutation(self) -> str:
        """
        Way of assigning the value to the gene

        Supported values:

        **"normal_distribution"** - assignes normally distributed random number to the variable selected for mutation

        **"random"** - Random values are assigned to genes
        """
        return self._gene_mutation

    @gene_mutation.setter
    def gene_mutation(self, value: str):
        self._gene_mutation = ga_utils.prepare_string(value)

    @property
    def cross_diversity_mutation_gene_sampling(self) -> str:
        """
        The sampling algorithm for mutating chromosomes when chromosome_mutation
        has value “cross_diversity”.
        It only applies when chromosome_mutation has value “cross_diversity” .
        It determines the way how genes are to be selected based on the cross-diversity.

        Supported values:

        **"roulette_wheel"** - The Roulette Wheel sampling algorithm
        (also known as “Weighted Random Pairing”). The probabilities
        assigned to the genes to be mutated are inversely
        proportional to their cross-diversity. A gene
        with the lowest cross-diversity has the greatest
        probability of mutation, while the gene with the
        highest cross-diversity has the lowest
        probability of mutation.

        **"tournament"** - The Tournament sampling algorithm. It randomly picks
        small subsets (groups) of genes, and genes with the lowest cross-diversity
        in subsets are chosen to be mutated. “tournament” can have
        an additional parameter separated from the “tournament”
        keyword by the comma. The other value represents a group size.
        For example, “tournament,3” means that the tournament
        mutation sampling algorithm is chosen, and each group
        contains up to 3 members. The default group size is 4.

        **"from_top_to_bottom"** - From Top To Bottom sampling
        algorithm starts at the top of the list and selects genes for mutation.

        **"random"** - Random sampling algorithm uses a
        uniform random number generator to select genes for mutation.
        In this case, sampling for the mutation will not
        depend on gene cross-diversity.
        """
        return self._cross_diversity_mutation_gene_sampling

    @cross_diversity_mutation_gene_sampling.setter
    def cross_diversity_mutation_gene_sampling(self, value: str):
        self._cross_diversity_mutation_gene_sampling = ga_utils.prepare_string(value)

    @property
    def max_attempt_no(self) -> int:
        """
        This parameter only takes place when exit_check has value “avg_cost”
        or “min_cost”. It determines the number of generations
        in which there is no improvement in the average/minimal cost.
        """
        return self._max_attempt_no

    @max_attempt_no.setter
    def max_attempt_no(self, value: int):
        self._max_attempt_no = ga_utils.try_get_int(value)

    @property
    def exit_check(self) -> str:
        """
        A criteria for the exit for the genetic algorithm.

        Supported values:

        **"avg_cost"** - The optimisation exit is triggered when the
        average cost of the upper half of the population
        is not improved in the specified number of generations

        **"min_cost"** - The optimisation exit is triggered when the minimal cost
        in the population is not improved in the specified number of generations

        **"requested"** - The optimisation exit is triggered when the
        requested value reached
        """
        return self._exit_check

    @exit_check.setter
    def exit_check(self, value: str):
        self._exit_check = ga_utils.prepare_string(value)

    @property
    def number_of_generations(self) -> float:
        """
        This parameter only takes place when exit_check
        has value “generations”. It determines the number of generations
        after which the genetic algorithm exits
        """
        return self._number_of_generations

    @number_of_generations.setter
    def number_of_generations(self, value: float):
        self._number_of_generations = ga_utils.try_get_float(value)

    @property
    def parent_selection(self) -> str:
        """
        The algorithm for sampling algorithm in parent selection.

        Supported values:

        **"roulette_wheel"** - Roulette Wheel sampling algorithm
        (also known as “Weighted Random Pairing”).
        The probabilities assigned to the chromosomes in the mating pool are
        inversely proportional to their cost. A chromosome with the lowest
        cost has the greatest probability of mating, while the chromosome
        with the highest cost has the lowest probability of mating.

        **"tournament"** - Tournament sampling algorithm. It randomly picks small
        subsets (groups) of chromosomes from the mating pool, and chromosomes
        with the lowest cost in subsets become a parent. “tournament”
        can have an additional parameter separated from the “tournament”
        keyword by the comma. The other value represents a group size.
        For example, “tournament,8” means that the tournament
        sampling algorithm for parent selection is chosen, and
        each group contains up to 8 members.
        The default group size is 4.

        **"from_top_to_bottom"** - From Top To Bottom sampling
        algorithm starts at the top of
        the list and pairs the chromosomes two at a time until the top kept chromosomes
        are selected for mating. Thus, the algorithm pairs odd rows with even rows.

        **"random"** - Random sampling algorithm uses a uniform random
        number generator to select chromosomes.
        """
        return self._parent_selection

    @parent_selection.setter
    def parent_selection(self, value: str):
        self._parent_selection = ga_utils.prepare_string(value)

    @property
    def crossover(self) -> str:
        """
        The crossover algorithm

        Supported values:

        **"blending"** - combines gene values from the two parents into new variable values in offsprings.
        One value of the offspring variable comes from a combination of the two
        corresponding values of the parental genes

        **"uniform"** - Genes from parents' chromosomes are combined in a uniform way.
        """
        return self._crossover

    @crossover.setter
    def crossover(self, value: str):
        self._crossover = ga_utils.prepare_string(value)

    @property
    def requested_cost(self) -> float:
        """
        This parameter only takes place when exit_check has value “requested”.
        It determines the requested value which causes the
        exit from the genetic algorithm
        """
        return self._requested_cost

    @requested_cost.setter
    def requested_cost(self, value: float):
        self._requested_cost = ga_utils.try_get_float(value)

    @property
    def logging(self) -> bool:
        return self._logging

    @logging.setter
    def logging(self, value: bool):
        """
        If this parameter has a True value, the log file will
        be created in the current working directory.
        The log file contains the flow of genetic algorithm execution,
        along with values of chromosomes, genes and
        cost functions in each generation
        """
        self._logging = ga_utils.try_get_bool(value)

    @property
    def must_mutate_for_same_parents(self) -> bool:
        """
        Indicates if completely the same parents must influence
        mutation for their children.
        In other words, each child will be mutated if it has parents
        with a diversity value of 0.
        If must_mutate_for_same_parents has the value True, the number of
        mutated chromosomes can outreach value determined
        by number_of_mutation_chromosomes
        or percentage_of_mutation_chromosomes
        """
        return self._must_mutate_for_same_parents

    @must_mutate_for_same_parents.setter
    def must_mutate_for_same_parents(self, value: bool):
        self._must_mutate_for_same_parents = ga_utils.try_get_bool(value)

    @property
    def timeout(self) -> int | str:
        """
        A number of seconds after which the genetic algorithm optimisation will\
            exit, regardless of whether exit_check criteria is reached.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value: int | str) -> None:
        self._timeout = ga_utils.try_get_int(value)

    def clone(self):
        # Create a new instance of the class
        new_instance = self.__class__.__new__(self.__class__)

        # Copy the attributes from the current instance to the new instance
        new_instance.__dict__ = copy.deepcopy(self.__dict__)

        return new_instance



import random
from typing import List

from gadapt.operations.mutation.chromosome_mutation.base_gene_mutation_rate_determinator import (
    BaseGeneMutationRateDeterminator,
)


class ComposedGeneMutationRateDeterminator(BaseGeneMutationRateDeterminator):
    """
    Provides a way to combine multiple gene mutation rate determinators into a single determinator. It randomly selects one of the determinators to determine the number of mutation genes in a chromosome.
    """

    def __init__(self) -> None:
        super().__init__()
        self.determinators: List[BaseGeneMutationRateDeterminator] = []

    def append(self, determinator: BaseGeneMutationRateDeterminator):
        """
        Adds a gene mutation rate determinator to the list of determinators.
        Args:
            determinator: A BaseGeneMutationRateDeterminator object representing a gene mutation rate determinator.
        """
        self.determinators.append(determinator)

    def _get_number_of_mutation_genes(self):
        if self.chromosome is None:
            raise Exception("Chromosome must not be null")
        if len(self.determinators) == 0:
            raise Exception("at least one mutator must be added")
        for determinator in self.determinators:
            determinator.chromosome = self.chromosome
            determinator.max_number_of_mutation_genes = (
                self.max_number_of_mutation_genes
            )
        if len(self.determinators) > 1:
            random.shuffle(self.determinators)
        current_determinator = self.determinators[0]
        return current_determinator._get_number_of_mutation_genes()



from abc import ABC, abstractmethod


class BaseChromosomeMutationRateDeterminator(ABC):
    """
    Provides a framework for determining the number of chromosomes to be mutated in a population.
    """

    def __init__(self) -> None:
        super().__init__()
        self.max_number_of_mutation_chromosomes = -1

    def get_number_of_mutation_chromosomes(
        self, population, max_number_of_mutation_chromosomes
    ):
        if population is None:
            raise Exception("Population must not be null")
        self.population = population
        self.max_number_of_mutation_chromosomes = max_number_of_mutation_chromosomes
        return self._get_number_of_mutation_chromosomes()

    @abstractmethod
    def _get_number_of_mutation_chromosomes(self) -> int:
        pass



import gadapt.utils.ga_utils as ga_utils
from gadapt.operations.mutation.chromosome_mutation.random_gene_mutation_rate_determinator import (
    RandomGeneMutationRateDeterminator,
)


class CrossDiversityGeneMutationRateDeterminator(RandomGeneMutationRateDeterminator):
    """
    Determines the number of mutation genes in a chromosome based on the cross diversity coefficient of the decision variables.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_number_of_mutation_genes(self) -> int:
        decision_variables = [g.decision_variable for g in self.chromosome]

        def get_mutation_rate() -> float:
            avg_rsd = ga_utils.average(
                [dv.cross_diversity_coefficient for dv in decision_variables]
            )
            if avg_rsd > 1:
                avg_rsd = 1
            if avg_rsd < 0:
                avg_rsd = 0
            return avg_rsd

        mutation_rate = get_mutation_rate()
        limit_number_of_mutation_genes = mutation_rate * float(
            self.max_number_of_mutation_genes
        )
        limit_number_of_mutation_genes_rounded = round(limit_number_of_mutation_genes)
        if limit_number_of_mutation_genes_rounded == 0:
            limit_number_of_mutation_genes_rounded = 1
        self.max_number_of_mutation_genes = limit_number_of_mutation_genes_rounded
        return super()._get_number_of_mutation_genes()



from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class StrictChromosomeMutationRateDeterminator(BaseChromosomeMutationRateDeterminator):
    """
    Provides a strict determination of the number of chromosomes to be mutated in a population.
    """

    def __init__(self) -> None:
        super().__init__()

    def _get_number_of_mutation_chromosomes(self) -> int:
        return self.max_number_of_mutation_chromosomes



from gadapt.operations.mutation.gene_mutation.base_gene_mutator import BaseGeneMutator


class RandomGeneMutator(BaseGeneMutator):
    """
    Generates a random value within the specified range of the decision variable.
    """

    def _make_mutated_value(self):
        return round(
            self.gene.decision_variable.make_random_value(),
            self.gene.decision_variable.decimal_places,
        )



import random

from gadapt.operations.mutation.gene_mutation.random_gene_mutator import (
    RandomGeneMutator,
)
from gadapt.utils.ga_utils import get_rand_bool, get_rand_bool_with_probability


# *** Obsolete ***
class ExtremePointedGeneMutator(RandomGeneMutator):

    def _make_mutated_value(self):
        return self._make_mutation()

    def _make_rounded_random_value_below_or_above(self):
        return round(
            self._make_random_value_below_or_above(),
            self.gene.decision_variable.decimal_places,
        )

    def _make_random_value_below_or_above(self):
        if get_rand_bool():
            return self._make_random_value_above()
        return self._make_random_value_below()

    def _make_random_value_below(self):
        if self.gene.variable_value == self.gene.decision_variable.min_value:
            return self.gene.decision_variable.make_random_value()
        number_of_steps = random.randint(
            0,
            round(
                (self.gene.variable_value - self.gene.decision_variable.min_value)
                / self.gene.decision_variable.step
            ),
        )
        return (
            self.gene.decision_variable.min_value
            + number_of_steps * self.gene.decision_variable.step
        )

    def _make_random_value_above(self):
        if self.gene.variable_value == self.gene.decision_variable.max_value:
            return self.gene.decision_variable.make_random_value()
        number_of_steps = random.randint(
            0,
            round(
                (self.gene.decision_variable.max_value - self.gene.variable_value)
                / self.gene.decision_variable.step
            ),
        )
        return (
            self.gene.variable_value
            + number_of_steps * self.gene.decision_variable.step
        )

    def _get_mutate_func(self):
        prob = self.gene.decision_variable.cross_diversity_coefficient
        if prob > 1.0:
            prob = 1.0
        should_mutate_random = get_rand_bool_with_probability(prob)
        if should_mutate_random:
            return super()._make_mutated_value
        else:
            return self._make_rounded_random_value_below_or_above

    def _make_mutation(self):
        f = self._get_mutate_func()
        return f()



import random
from typing import List

from gadapt.operations.sampling.base_sampling import T, BaseSampling


class TournamentSampling(BaseSampling):
    """
    "Tournament" algorithm for extracting a sample from the population.
    """

    def __init__(self, group_size=None) -> None:
        super().__init__()
        self.group_size = group_size

    def _prepare_sample(self, lst: List[T]) -> List[T]:
        ls = []
        ls = sorted(lst, key=self._sort_key)
        groups = self._get_groups(ls)
        self._play_tournament(groups, self._sort_key)
        return self._make_ranking(groups)

    def _get_groups(self, lst: List[T]) -> List[List[T]]:
        size = len(lst)
        ls = [rm for rm in lst]
        if self.group_size is None or self.group_size > len(ls):
            self.group_size = self._calculate_group_size(ls)
        num_of_groups = size // self.group_size
        # if size % self.group_size > 0:
        #    num_of_groups += 1
        groups = []
        for _ in range(num_of_groups):
            l_groups: List[T] = []
            groups.append(l_groups)
        for i in range(self.group_size):
            for g in groups:
                random_index = random.randint(0, len(ls) - 1)
                element = ls.pop(random_index)
                g.append(element)
                if len(ls) == 0:
                    break
            if len(ls) == 0:
                break
        if len(ls) > 0:
            groups.append(ls)
        return groups

    def _calculate_group_size(self, ls: List[T]):
        size = len(ls)
        if size <= 3:
            return len(ls)
        group_size = 2
        if size >= 9:
            group_size += 1
        if size >= 12:
            group_size += 1
        return group_size

    def _play_tournament(self, groups: List[List[T]], sort_key):
        for g in groups:
            g.sort(key=sort_key)

    def _make_ranking(self, groups: List[List[T]]):
        rank = 0
        members_for_action = []
        number_of_groups = len(groups)
        while rank < self.max_num:
            number_of_empty_groups = 0
            for g in groups:
                if len(g) == 0:
                    number_of_empty_groups += 1
                    continue
                element = g.pop(0)
                element.rank = rank
                members_for_action.append(element)
                rank += 1
            if number_of_empty_groups == number_of_groups:
                break
        return members_for_action[: self.max_num]



import datetime
import logging


class TimestampFormatter(logging.Formatter):
    """
    Formatter for timestamps
    """

    def format(self, record):
        """
        Format the specified record as text with the provided date time format.
        Args:
            record: Record to format
        """
        record.asctime = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        return super().format(record)



import math

from gadapt.operations.population_update.base_population_updater import (
    BasePopulationUpdater,
)
from gadapt.utils import ga_utils


class CommonPopulationUpdater(BasePopulationUpdater):
    """
    Common population updater
    """

    def _calculate_average_cost_step(self):
        allocated_values = [
            c.cost_value
            for c in self.population.chromosomes
            if c.cost_value is not None and not math.isnan(c.cost_value)
        ]
        if allocated_values:
            return ga_utils.average_difference(allocated_values)
        return float("NaN")

    def _update_population(self):
        self.population.average_cost_step = self._calculate_average_cost_step()
        if math.isnan(self.population.average_cost_step_in_first_population):
            self.population.average_cost_step_in_first_population = (
                self.population.average_cost_step
            )



"""
Population
"""

import sys
from datetime import datetime
from typing import List

import gadapt.adapters.string_operation.ga_strings as ga_strings
import gadapt.ga_model.definitions as definitions
from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.ga_options import GAOptions
from gadapt.ga_model.gene import Gene


class Population:
    def __init__(self, options: GAOptions):
        """Population for the genetic algorithm. It contains a collection of\
            chromosomes, as well as additional parameters

        Args:

            options (GAOptions): Genetic Algorithm Options
        """
        self._previous_min_cost = sys.float_info.min
        self._avg_cost = sys.float_info.min
        self._min_cost = sys.float_info.min
        if __name__ == "__main__":
            self._previous_avg_cost = sys.float_info.min
        if options.population_size < 4:
            raise Exception("Population size 4 must be higher than 3")
        self.options = options
        self._set_init_values()
        self.last_chromosome_id = 1
        self._population_generation = 0
        self.options = options
        self.chromosomes: List[Chromosome] = []
        self.generate_initial_population()
        self.start_time = datetime.now()
        self.average_cost_step = float("NaN")
        self.average_cost_step_in_first_population = float("NaN")
        self.timeout_expired = False
        self.min_cost_per_generation = []

    def __iter__(self):
        return PopulationIterator(self)

    def __getitem__(self, index):
        return self.chromosomes[index]

    def __next__(self):
        return next(self.chromosomes)

    def __len__(self):
        return len(self.chromosomes)

    def __str__(self):
        return self._to_string()

    def get_sorted(self, key=None, reverse: bool = False):
        """Sorted list of chromosomes
        Args:
            key: Sorted key
            reverse (bool=False): is reversed
        """
        return sorted(self.chromosomes, key=key, reverse=reverse)

    def append(self, c: Chromosome):
        self.chromosomes.append(c)

    def generate_initial_population(self):
        for i in range(self.options.population_size):
            self.add_new_chromosome()

    def _to_string(self):
        return ga_strings.population_to_string(self)

    def _set_init_values(self):
        float_init_value = definitions.FLOAT_NAN
        self.avg_cost = float_init_value
        self.previous_avg_cost = float_init_value
        self.min_cost = float_init_value
        self.previous_min_cost = float_init_value

    @property
    def options(self) -> GAOptions:
        """
        Genetic algorithm options
        """
        return self._options

    @options.setter
    def options(self, value: GAOptions):
        self._options = value

    @property
    def avg_cost(self) -> float:
        """
        Average cost of the population
        """
        return self._avg_cost

    @avg_cost.setter
    def avg_cost(self, value: float):
        self._avg_cost = value

    @property
    def previous_avg_cost(self) -> float:
        """
        Previous average cost
        """
        return self._previous_avg_cost

    @previous_avg_cost.setter
    def previous_avg_cost(self, value: float):
        self._previous_avg_cost = value

    @property
    def min_cost(self):
        """
        Minimum cost
        """
        return self._min_cost

    @min_cost.setter
    def min_cost(self, value: float):
        self._min_cost = value

    @property
    def previous_min_cost(self):
        """
        Previous minimum cost
        """
        return self._previous_min_cost

    @previous_min_cost.setter
    def previous_min_cost(self, value: float):
        self._previous_min_cost = value

    @property
    def best_individual(self) -> Chromosome:
        """
        Best individual chromosome
        """
        return self._best_individual

    @best_individual.setter
    def best_individual(self, value: Chromosome):
        self._best_individual = value

    @property
    def population_generation(self):
        """
        Current generation of the population
        """
        return self._population_generation

    @population_generation.setter
    def population_generation(self, value):
        self._population_generation = value

    def clear(self):
        """
        Clears all chromosomes
        """
        self.chromosomes.clear()

    def clear_and_add_chromosomes(self, chromosomes: List[Chromosome]):
        """
        Clears chromosomes and adds new ones
        Args:
            chromosomes (List[Chromosome]): chromosomes to add
        """
        self.chromosomes.clear()
        self.add_chromosomes(chromosomes)

    def add_chromosomes(self, chromosomes):
        """
        Adds chromosomes to population
        Args:
            chromosomes (Tuple[Chromosome]): chromosomes to add
        """
        for c in chromosomes:
            self.add_chromosome(c)

    def add_new_chromosome(self):
        """
        Adds new chromosomes to the population
        """
        chromosome = Chromosome(
            self.population_generation,
        )
        chromosome.chromosome_generation = 1
        self.add_chromosome(chromosome)

    def add_chromosome(self, chromosome):
        """
        Adds chromosome to the population
        Args:
            chromosome: chromosome to add
        """
        if len(self) >= self.options.population_size:
            return
        if chromosome.chromosome_id is None or chromosome.chromosome_id == -1:
            chromosome.chromosome_id = self.last_chromosome_id
            self.last_chromosome_id += 1
        if len(chromosome) == 0:
            for dv in self.options.decision_variables:
                g = Gene(dv)
                chromosome.append(g)
        self.append(chromosome)

    def update_variables(self):
        """
        Updates decision variables
        """
        self.variable_updater.update_variables(self)

    def update_population(self):
        """
        Updates population
        """
        self.population_updater.update_population(self)


class PopulationIterator:
    def __init__(self, population):
        self.population = population
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.population.chromosomes):
            result = self.population.chromosomes[self.index]
            self.index += 1
            return result
        else:
            raise StopIteration



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class AvgCostExitChecker(BaseExitChecker):
    """
    Exit check based on average cost.
    The GA exits when there is no improvement in
    the average cost of the kept population in a defined number of iterations.
    """

    def _is_exit(self):
        if self.population is None:
            raise Exception("population must not be null")
        return self.population.avg_cost >= self.population.previous_avg_cost



from typing import List

from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.immigration.population_immigration.base_population_immigrator import (
    BasePopulationImmigrator,
)


class CommonPopulationImmigrator(BasePopulationImmigrator):
    """
    Common class for the self.population immigration.
    In kept part of the self.population lower ranked chromosomes are replaced with new ones
    """

    def _immigrate_population(self):
        if self.population.options.immigration_number < 1:
            return
        keep_number = self.population.options.keep_number
        chromosome_list: List[Chromosome] = list(
            self.population.get_sorted(key=lambda c: c.cost_value)
        )[:keep_number]
        chromosome_list = sorted(
            chromosome_list, key=lambda c: (-c.cost_value, -c.chromosome_id)
        )[: self.population.options.immigration_number]
        for c in chromosome_list:
            self._chromosome_immigrator.immigrate(c)
            c.population_generation = self.population.population_generation



from setuptools import setup, find_packages

setup(
    name="gadapt",
    version="0.4.11",
    author="Zoran Jankovic",
    author_email="bpzoran@yahoo.com",
    url="https://github.com/bpzoran/gadapt",
    packages=find_packages(),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    description="GAdapt: A Python Library for\
        Self-Adaptive Genetic Algorithm.",
    install_requires=["numpy==1.26.4"],
)



"""
Ranking model
"""

import gadapt.ga_model.definitions as definitions


class RankingModel:
    def __init__(self):
        """
        Base class for the object that can be ranked\
            (chromosomes and genes), for sampling purposes.
        """
        self._rank = -1
        self._cummulative_probability = definitions.FLOAT_NAN

    @property
    def rank(self):
        """
        Rank of the object
        """
        return self._rank

    @rank.setter
    def rank(self, value):
        self._rank = value

    @property
    def action_probability(self):
        """
        A probability for executing some action (e.g. pairing or mutation)
        """
        return self._cummulative_probability

    @action_probability.setter
    def action_probability(self, value):
        self._cummulative_probability = value

    def set_rank_and_return(self, rank):
        """
        Sets rank for the object and returns the object
        """
        self.rank = rank
        return self

    def reset_for_sampling(self):
        """
        Resets object properties for the purpose of a new sampling
        """
        self._rank = -1
        self._cummulative_probability = definitions.FLOAT_NAN
